import shutil
import tempfile

from unittest import TestCase
from mock import Mock

from autosubmit.job.job_common import Status
from autosubmit.job.job_list import JobList
from autosubmit.job.job_list_persistence import JobListPersistenceDb
from autosubmitconfigparser.config.yamlparser import YAMLParserFactory
from random import randrange
from autosubmit.job.job import Job
from autosubmit.monitor.monitor import Monitor

class TestJobGraph(TestCase):

    def setUp(self):
        self.experiment_id = 'random-id'
        self.as_conf = Mock()
        self.as_conf.experiment_data = dict()
        self.as_conf.experiment_data["JOBS"] = dict()
        self.as_conf.jobs_data = self.as_conf.experiment_data["JOBS"]
        self.as_conf.experiment_data["PLATFORMS"] = dict()
        self.temp_directory = tempfile.mkdtemp()
        self.job_list = JobList(self.experiment_id, FakeBasicConfig, YAMLParserFactory(),
                                JobListPersistenceDb(self.temp_directory, 'db'),self.as_conf)
        self.parser_mock = Mock(spec='SafeConfigParser')

        # Basic workflow with SETUP, INI, SIM, POST, CLEAN
        setup_job = self._createDummyJob('expid_SETUP', Status.READY)
        self.job_list.get_job_list().append(setup_job)

        for date in ['d1', 'd2']:
            for member in ['m1', 'm2']:
                job = self._createDummyJob('expid_' + date + '_' + member + '_' + 'INI', Status.WAITING, date, member)
                job.add_parent(setup_job)
                self.job_list.get_job_list().append(job)

        sections = ['SIM', 'POST', 'CLEAN']
        for section in sections:
            for date in ['d1', 'd2']:
                for member in ['m1', 'm2']:
                    for chunk in [1, 2]:
                        job = self._createDummyJob('expid_' + date + '_' + member + '_' + str(chunk) + '_' + section,
                                                   Status.WAITING, date, member, chunk)
                        if section == 'SIM':
                            if chunk > 1:
                                job.add_parent(self.job_list.get_job_by_name('expid_'+date+'_'+member+'_'+str(chunk-1)+'_SIM'))
                            else:
                                job.add_parent(self.job_list.get_job_by_name('expid_'+date+'_'+member+'_INI'))
                        elif section == 'POST':
                            job.add_parent(self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))
                        elif section == 'CLEAN':
                            job.add_parent(self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_POST'))
                        self.job_list.get_job_list().append(job)

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_directory)

    def test_grouping_date(self):
        groups_dict = dict()
        groups_dict['status'] = {'d1': Status.WAITING, 'd2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_INI' : ['d1'], 'expid_d1_m2_INI' : ['d1'], 'expid_d2_m1_INI' : ['d2'], 'expid_d2_m2_INI' : ['d2'],
            'expid_d1_m1_1_SIM': ['d1'], 'expid_d1_m1_2_SIM': ['d1'], 'expid_d1_m2_1_SIM': ['d1'],
            'expid_d1_m2_2_SIM': ['d1'],
            'expid_d2_m1_1_SIM': ['d2'], 'expid_d2_m1_2_SIM': ['d2'], 'expid_d2_m2_1_SIM': ['d2'],
            'expid_d2_m2_2_SIM': ['d2'],

            'expid_d1_m1_1_POST': ['d1'], 'expid_d1_m1_2_POST': ['d1'], 'expid_d1_m2_1_POST': ['d1'],
            'expid_d1_m2_2_POST': ['d1'],
            'expid_d2_m1_1_POST': ['d2'], 'expid_d2_m1_2_POST': ['d2'], 'expid_d2_m2_1_POST': ['d2'],
            'expid_d2_m2_2_POST': ['d2'],

            'expid_d1_m1_1_CLEAN': ['d1'], 'expid_d1_m1_2_CLEAN': ['d1'], 'expid_d1_m2_1_CLEAN': ['d1'],
            'expid_d1_m2_2_CLEAN': ['d1'],
            'expid_d2_m1_1_CLEAN': ['d2'], 'expid_d2_m1_2_CLEAN': ['d2'], 'expid_d2_m2_1_CLEAN': ['d2'],
            'expid_d2_m2_2_CLEAN': ['d2']
        }

        nodes = [
            "expid_SETUP", 'd1', 'd2'
        ]
        edges = [
            ("expid_SETUP", "d1"), ("expid_SETUP", "d2"), ("d1", "d1"), ("d2", "d2")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict)

        self.assertTrue(graph.obj_dict['strict'])

        subgraph = graph.obj_dict['subgraphs']['Experiment'][0]

        self.assertListEqual(sorted(list(subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(subgraph['edges'].keys())), sorted(edges))

    def test_grouping_member(self):
        groups_dict = dict()
        groups_dict['status'] = {'d1_m1': Status.WAITING,
                                 'd1_m2': Status.WAITING,
                                 'd2_m1': Status.WAITING,
                                 'd2_m2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_INI': ['d1_m1'], 'expid_d1_m2_INI': ['d1_m2'], 'expid_d2_m1_INI': ['d2_m1'], 'expid_d2_m2_INI': ['d2_m2'],
            'expid_d1_m1_1_SIM': ['d1_m1'], 'expid_d1_m1_2_SIM': ['d1_m1'], 'expid_d1_m2_1_SIM': ['d1_m2'],
            'expid_d1_m2_2_SIM': ['d1_m2'],
            'expid_d2_m1_1_SIM': ['d2_m1'], 'expid_d2_m1_2_SIM': ['d2_m1'], 'expid_d2_m2_1_SIM': ['d2_m2'],
            'expid_d2_m2_2_SIM': ['d2_m2'],

            'expid_d1_m1_1_POST': ['d1_m1'], 'expid_d1_m1_2_POST': ['d1_m1'], 'expid_d1_m2_1_POST': ['d1_m2'],
            'expid_d1_m2_2_POST': ['d1_m2'],
            'expid_d2_m1_1_POST': ['d2_m1'], 'expid_d2_m1_2_POST': ['d2_m1'], 'expid_d2_m2_1_POST': ['d2_m2'],
            'expid_d2_m2_2_POST': ['d2_m2'],

            'expid_d1_m1_1_CLEAN': ['d1_m1'], 'expid_d1_m1_2_CLEAN': ['d1_m1'], 'expid_d1_m2_1_CLEAN': ['d1_m2'],
            'expid_d1_m2_2_CLEAN': ['d1_m2'],
            'expid_d2_m1_1_CLEAN': ['d2_m1'], 'expid_d2_m1_2_CLEAN': ['d2_m1'], 'expid_d2_m2_1_CLEAN': ['d2_m2'],
            'expid_d2_m2_2_CLEAN': ['d2_m2']
        }

        nodes = [
            "expid_SETUP",
            'd1_m1', 'd1_m2', 'd2_m2', 'd2_m1'
        ]
        edges = [
            ("expid_SETUP", "d1_m1"), ("expid_SETUP", "d1_m2"), ("expid_SETUP", "d2_m1"),
            ("expid_SETUP", "d2_m2"),

            ("d1_m1", "d1_m1"), ("d1_m2", "d1_m2"),
            ("d2_m1", "d2_m1"), ("d2_m2", "d2_m2")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict)

        self.assertTrue(graph.obj_dict['strict'])

        subgraph = graph.obj_dict['subgraphs']['Experiment'][0]

        self.assertListEqual(sorted(list(subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(subgraph['edges'].keys())), sorted(edges))

    def test_grouping_chunk(self):
        groups_dict = dict()
        groups_dict['status'] = {'d1_m1_1': Status.WAITING, 'd1_m1_2': Status.WAITING,
                                 'd1_m2_1': Status.WAITING, 'd1_m2_2': Status.WAITING,
                                 'd2_m1_1': Status.WAITING, 'd2_m1_2': Status.WAITING,
                                 'd2_m2_1': Status.WAITING, 'd2_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_1_SIM': ['d1_m1_1'], 'expid_d1_m1_2_SIM': ['d1_m1_2'], 'expid_d1_m2_1_SIM': ['d1_m2_1'],
            'expid_d1_m2_2_SIM': ['d1_m2_2'],
            'expid_d2_m1_1_SIM': ['d2_m1_1'], 'expid_d2_m1_2_SIM': ['d2_m1_2'], 'expid_d2_m2_1_SIM': ['d2_m2_1'],
            'expid_d2_m2_2_SIM': ['d2_m2_2'],

            'expid_d1_m1_1_POST': ['d1_m1_1'], 'expid_d1_m1_2_POST': ['d1_m1_2'], 'expid_d1_m2_1_POST': ['d1_m2_1'],
            'expid_d1_m2_2_POST': ['d1_m2_2'],
            'expid_d2_m1_1_POST': ['d2_m1_1'], 'expid_d2_m1_2_POST': ['d2_m1_2'], 'expid_d2_m2_1_POST': ['d2_m2_1'],
            'expid_d2_m2_2_POST': ['d2_m2_2'],

            'expid_d1_m1_1_CLEAN': ['d1_m1_1'], 'expid_d1_m1_2_CLEAN': ['d1_m1_2'], 'expid_d1_m2_1_CLEAN': ['d1_m2_1'],
            'expid_d1_m2_2_CLEAN': ['d1_m2_2'],
            'expid_d2_m1_1_CLEAN': ['d2_m1_1'], 'expid_d2_m1_2_CLEAN': ['d2_m1_2'], 'expid_d2_m2_1_CLEAN': ['d2_m2_1'],
            'expid_d2_m2_2_CLEAN': ['d2_m2_2']
        }

        nodes = [
            "expid_SETUP", "expid_d1_m1_INI", "expid_d1_m2_INI", "expid_d2_m1_INI", "expid_d2_m2_INI",
            'd1_m1_1', 'd1_m2_1', 'd2_m1_1', 'd2_m2_1', 'd1_m1_2', 'd1_m2_2', 'd2_m1_2', 'd2_m2_2'
        ]
        edges = [
            ("expid_SETUP", "expid_d1_m1_INI"), ("expid_SETUP", "expid_d1_m2_INI"), ("expid_SETUP", "expid_d2_m1_INI"),
            ("expid_SETUP", "expid_d2_m2_INI"), ("expid_d1_m1_INI", "d1_m1_1"), ("expid_d1_m2_INI", "d1_m2_1"),
            ("expid_d2_m1_INI", "d2_m1_1"), ("expid_d2_m2_INI", "d2_m2_1"),
            ("d1_m1_1", "d1_m1_2"), ("d1_m2_1", "d1_m2_2"), ("d2_m1_1", "d2_m1_2"), ("d2_m2_1", "d2_m2_2"),

            ("d1_m1_1", "d1_m1_1"), ("d1_m1_2", "d1_m1_2"), ("d1_m2_1", "d1_m2_1"), ("d1_m2_2", "d1_m2_2"),
            ("d2_m1_1", "d2_m1_1"), ("d2_m1_2", "d2_m1_2"), ("d2_m2_1", "d2_m2_1"), ("d2_m2_2", "d2_m2_2")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict)

        self.assertTrue(graph.obj_dict['strict'])

        subgraph = graph.obj_dict['subgraphs']['Experiment'][0]

        self.assertListEqual(sorted(list(subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(subgraph['edges'].keys())), sorted(edges))

    def test_grouping_automatic_hide(self):
        groups_dict = dict()
        groups_dict['status'] = {'d1_m1_1': Status.WAITING, 'd1_m1_2': Status.WAITING,
                                 'd1_m2_1': Status.WAITING, 'd1_m2_2': Status.WAITING,
                                 'd2_m1_1': Status.WAITING, 'd2_m1_2': Status.WAITING,
                                 'd2_m2_1': Status.WAITING, 'd2_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_1_SIM': ['d1_m1_1'], 'expid_d1_m1_2_SIM': ['d1_m1_2'], 'expid_d1_m2_1_SIM': ['d1_m2_1'],
            'expid_d1_m2_2_SIM': ['d1_m2_2'],
            'expid_d2_m1_1_SIM': ['d2_m1_1'], 'expid_d2_m1_2_SIM': ['d2_m1_2'], 'expid_d2_m2_1_SIM': ['d2_m2_1'],
            'expid_d2_m2_2_SIM': ['d2_m2_2'],

            'expid_d1_m1_1_POST': ['d1_m1_1'], 'expid_d1_m1_2_POST': ['d1_m1_2'], 'expid_d1_m2_1_POST': ['d1_m2_1'],
            'expid_d1_m2_2_POST': ['d1_m2_2'],
            'expid_d2_m1_1_POST': ['d2_m1_1'], 'expid_d2_m1_2_POST': ['d2_m1_2'], 'expid_d2_m2_1_POST': ['d2_m2_1'],
            'expid_d2_m2_2_POST': ['d2_m2_2'],

            'expid_d1_m1_1_CLEAN': ['d1_m1_1'], 'expid_d1_m1_2_CLEAN': ['d1_m1_2'], 'expid_d1_m2_1_CLEAN': ['d1_m2_1'],
            'expid_d1_m2_2_CLEAN': ['d1_m2_2'],
            'expid_d2_m1_1_CLEAN': ['d2_m1_1'], 'expid_d2_m1_2_CLEAN': ['d2_m1_2'], 'expid_d2_m2_1_CLEAN': ['d2_m2_1'],
            'expid_d2_m2_2_CLEAN': ['d2_m2_2']
        }

        nodes = [
            "expid_SETUP", "expid_d1_m1_INI", "expid_d1_m2_INI", "expid_d2_m1_INI", "expid_d2_m2_INI"
        ]
        edges = [
            ("expid_SETUP", "expid_d1_m1_INI"), ("expid_SETUP", "expid_d1_m2_INI"), ("expid_SETUP", "expid_d2_m1_INI"),
            ("expid_SETUP", "expid_d2_m2_INI")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict, hide_groups=True)

        self.assertTrue(graph.obj_dict['strict'])

        subgraph = graph.obj_dict['subgraphs']['Experiment'][0]

        self.assertListEqual(sorted(list(subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(subgraph['edges'].keys())), sorted(edges))

    def test_synchronize_member(self):
        for date in ['d1', 'd2']:
            for chunk in [1, 2]:
                job = self._createDummyJob('expid_' + date + '_' + str(chunk) + '_ASIM',
                                           Status.WAITING, date, None, chunk)
                for member in ['m1', 'm2']:
                    job.add_parent(self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))
                self.job_list.get_job_list().append(job)

        nodes = [
            "expid_SETUP", "expid_d1_m1_INI", "expid_d1_m2_INI", "expid_d2_m1_INI", "expid_d2_m2_INI",
            "expid_d1_m1_1_SIM", "expid_d1_m1_2_SIM", "expid_d1_m2_1_SIM", "expid_d1_m2_2_SIM",
            "expid_d2_m1_1_SIM", "expid_d2_m1_2_SIM", "expid_d2_m2_1_SIM", "expid_d2_m2_2_SIM",
            "expid_d1_m1_1_POST", "expid_d1_m1_2_POST", "expid_d1_m2_1_POST", "expid_d1_m2_2_POST",
            "expid_d2_m1_1_POST", "expid_d2_m1_2_POST", "expid_d2_m2_1_POST", "expid_d2_m2_2_POST",
            "expid_d1_m1_1_CLEAN", "expid_d1_m1_2_CLEAN", "expid_d1_m2_1_CLEAN", "expid_d1_m2_2_CLEAN",
            "expid_d2_m1_1_CLEAN", "expid_d2_m1_2_CLEAN", "expid_d2_m2_1_CLEAN", "expid_d2_m2_2_CLEAN",
            "expid_d1_1_ASIM", "expid_d1_2_ASIM", "expid_d2_1_ASIM", "expid_d2_2_ASIM"
        ]
        edges = [
            ('expid_SETUP', 'expid_d1_m1_INI'), ('expid_SETUP', 'expid_d1_m2_INI'), ('expid_SETUP', 'expid_d2_m1_INI'),
            ('expid_SETUP', 'expid_d2_m2_INI'),

            ('expid_d1_m1_INI', 'expid_d1_m1_1_SIM'), ('expid_d1_m2_INI', 'expid_d1_m2_1_SIM'),
            ('expid_d2_m1_INI', 'expid_d2_m1_1_SIM'), ('expid_d2_m2_INI', 'expid_d2_m2_1_SIM'),
            
            ('expid_d1_m1_1_SIM', 'expid_d1_m1_2_SIM'),
            ('expid_d1_m1_1_SIM', 'expid_d1_m1_1_POST'),
            ('expid_d1_m1_1_SIM', 'expid_d1_1_ASIM'),

            ('expid_d1_m1_2_SIM', 'expid_d1_m1_2_POST'),
            ('expid_d1_m1_2_SIM', 'expid_d1_2_ASIM'),

            ('expid_d1_m2_1_SIM', 'expid_d1_m2_2_SIM'),
            ('expid_d1_m2_1_SIM', 'expid_d1_m2_1_POST'),
            ('expid_d1_m2_1_SIM', 'expid_d1_1_ASIM'),

            ('expid_d1_m2_2_SIM', 'expid_d1_m2_2_POST'),
            ('expid_d1_m2_2_SIM', 'expid_d1_2_ASIM'),

            ('expid_d2_m1_1_SIM', 'expid_d2_m1_2_SIM'),
            ('expid_d2_m1_1_SIM', 'expid_d2_m1_1_POST'),
            ('expid_d2_m1_1_SIM', 'expid_d2_1_ASIM'),

            ('expid_d2_m1_2_SIM', 'expid_d2_m1_2_POST'),
            ('expid_d2_m1_2_SIM', 'expid_d2_2_ASIM'),

            ('expid_d2_m2_1_SIM', 'expid_d2_m2_2_SIM'),
            ('expid_d2_m2_1_SIM', 'expid_d2_m2_1_POST'),
            ('expid_d2_m2_1_SIM', 'expid_d2_1_ASIM'),

            ('expid_d2_m2_2_SIM', 'expid_d2_m2_2_POST'),
            ('expid_d2_m2_2_SIM', 'expid_d2_2_ASIM'),

            ('expid_d1_m1_1_POST', 'expid_d1_m1_1_CLEAN'),
            ('expid_d1_m1_2_POST', 'expid_d1_m1_2_CLEAN'),
            ('expid_d1_m2_1_POST', 'expid_d1_m2_1_CLEAN'),
            ('expid_d1_m2_2_POST', 'expid_d1_m2_2_CLEAN'),

            ('expid_d2_m1_1_POST', 'expid_d2_m1_1_CLEAN'),
            ('expid_d2_m1_2_POST', 'expid_d2_m1_2_CLEAN'),
            ('expid_d2_m2_1_POST', 'expid_d2_m2_1_CLEAN'),
            ('expid_d2_m2_2_POST', 'expid_d2_m2_2_CLEAN')
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, dict())

        self.assertFalse(graph.obj_dict['strict'])

        subgraph = graph.obj_dict['subgraphs']['Experiment'][0]

        self.assertListEqual(sorted(list(subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(subgraph['edges'].keys())), sorted(edges))

    def test_synchronize_date(self):
        for chunk in [1, 2]:
            job = self._createDummyJob('expid_' + str(chunk) + '_ASIM',
                                       Status.WAITING, None, None, chunk)
            for date in ['d1', 'd2']:
                for member in ['m1', 'm2']:
                    job.add_parent(self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))
            self.job_list.get_job_list().append(job)

        nodes = [
            "expid_SETUP", "expid_d1_m1_INI", "expid_d1_m2_INI", "expid_d2_m1_INI", "expid_d2_m2_INI",
            "expid_d1_m1_1_SIM", "expid_d1_m1_2_SIM", "expid_d1_m2_1_SIM", "expid_d1_m2_2_SIM",
            "expid_d2_m1_1_SIM", "expid_d2_m1_2_SIM", "expid_d2_m2_1_SIM", "expid_d2_m2_2_SIM",
            "expid_d1_m1_1_POST", "expid_d1_m1_2_POST", "expid_d1_m2_1_POST", "expid_d1_m2_2_POST",
            "expid_d2_m1_1_POST", "expid_d2_m1_2_POST", "expid_d2_m2_1_POST", "expid_d2_m2_2_POST",
            "expid_d1_m1_1_CLEAN", "expid_d1_m1_2_CLEAN", "expid_d1_m2_1_CLEAN", "expid_d1_m2_2_CLEAN",
            "expid_d2_m1_1_CLEAN", "expid_d2_m1_2_CLEAN", "expid_d2_m2_1_CLEAN", "expid_d2_m2_2_CLEAN",
            "expid_1_ASIM", "expid_2_ASIM"
        ]
        edges = [
            ('expid_SETUP', 'expid_d1_m1_INI'), ('expid_SETUP', 'expid_d1_m2_INI'), ('expid_SETUP', 'expid_d2_m1_INI'),
            ('expid_SETUP', 'expid_d2_m2_INI'),

            ('expid_d1_m1_INI', 'expid_d1_m1_1_SIM'), ('expid_d1_m2_INI', 'expid_d1_m2_1_SIM'),
            ('expid_d2_m1_INI', 'expid_d2_m1_1_SIM'), ('expid_d2_m2_INI', 'expid_d2_m2_1_SIM'),

            ('expid_d1_m1_1_SIM', 'expid_d1_m1_2_SIM'),
            ('expid_d1_m1_1_SIM', 'expid_d1_m1_1_POST'),
            ('expid_d1_m1_1_SIM', 'expid_1_ASIM'),

            ('expid_d1_m1_2_SIM', 'expid_d1_m1_2_POST'),
            ('expid_d1_m1_2_SIM', 'expid_2_ASIM'),

            ('expid_d1_m2_1_SIM', 'expid_d1_m2_2_SIM'),
            ('expid_d1_m2_1_SIM', 'expid_d1_m2_1_POST'),
            ('expid_d1_m2_1_SIM', 'expid_1_ASIM'),

            ('expid_d1_m2_2_SIM', 'expid_d1_m2_2_POST'),
            ('expid_d1_m2_2_SIM', 'expid_2_ASIM'),

            ('expid_d2_m1_1_SIM', 'expid_d2_m1_2_SIM'),
            ('expid_d2_m1_1_SIM', 'expid_d2_m1_1_POST'),
            ('expid_d2_m1_1_SIM', 'expid_1_ASIM'),

            ('expid_d2_m1_2_SIM', 'expid_d2_m1_2_POST'),
            ('expid_d2_m1_2_SIM', 'expid_2_ASIM'),

            ('expid_d2_m2_1_SIM', 'expid_d2_m2_2_SIM'),
            ('expid_d2_m2_1_SIM', 'expid_d2_m2_1_POST'),
            ('expid_d2_m2_1_SIM', 'expid_1_ASIM'),

            ('expid_d2_m2_2_SIM', 'expid_d2_m2_2_POST'),
            ('expid_d2_m2_2_SIM', 'expid_2_ASIM'),

            ('expid_d1_m1_1_POST', 'expid_d1_m1_1_CLEAN'),
            ('expid_d1_m1_2_POST', 'expid_d1_m1_2_CLEAN'),
            ('expid_d1_m2_1_POST', 'expid_d1_m2_1_CLEAN'),
            ('expid_d1_m2_2_POST', 'expid_d1_m2_2_CLEAN'),

            ('expid_d2_m1_1_POST', 'expid_d2_m1_1_CLEAN'),
            ('expid_d2_m1_2_POST', 'expid_d2_m1_2_CLEAN'),
            ('expid_d2_m2_1_POST', 'expid_d2_m2_1_CLEAN'),
            ('expid_d2_m2_2_POST', 'expid_d2_m2_2_CLEAN')
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, dict())

        self.assertFalse(graph.obj_dict['strict'])

        subgraph = graph.obj_dict['subgraphs']['Experiment'][0]

        self.assertListEqual(sorted(list(subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(subgraph['edges'].keys())), sorted(edges))

    def test_wrapper_package(self):
        packages = [('expid', 'package_d1_m1_SIM', 'expid_d1_m1_1_SIM'), ('expid', 'package_d1_m1_SIM', 'expid_d1_m1_2_SIM'),
                    ('expid', 'package_d2_m2_SIM', 'expid_d2_m2_1_SIM'), ('expid', 'package_d2_m2_SIM', 'expid_d2_m2_2_SIM')]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), packages, dict())
        self.assertFalse(graph.obj_dict['strict'])
        for (expid, package, job_name) in packages:
            self.assertIn('cluster_'+package, graph.obj_dict['subgraphs'])

    def test_synchronize_member_group_member(self):
        for date in ['d1', 'd2']:
            for chunk in [1, 2]:
                job = self._createDummyJob('expid_' + date + '_' + str(chunk) + '_ASIM',
                                           Status.WAITING, date, None, chunk)
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

                self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'d1_m1': Status.WAITING,
                                 'd1_m2': Status.WAITING,
                                 'd2_m1': Status.WAITING,
                                 'd2_m2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_INI': ['d1_m1'], 'expid_d1_m2_INI': ['d1_m2'], 'expid_d2_m1_INI': ['d2_m1'],
            'expid_d2_m2_INI': ['d2_m2'],
            'expid_d1_m1_1_SIM': ['d1_m1'], 'expid_d1_m1_2_SIM': ['d1_m1'], 'expid_d1_m2_1_SIM': ['d1_m2'],
            'expid_d1_m2_2_SIM': ['d1_m2'],
            'expid_d2_m1_1_SIM': ['d2_m1'], 'expid_d2_m1_2_SIM': ['d2_m1'], 'expid_d2_m2_1_SIM': ['d2_m2'],
            'expid_d2_m2_2_SIM': ['d2_m2'],

            'expid_d1_m1_1_POST': ['d1_m1'], 'expid_d1_m1_2_POST': ['d1_m1'], 'expid_d1_m2_1_POST': ['d1_m2'],
            'expid_d1_m2_2_POST': ['d1_m2'],
            'expid_d2_m1_1_POST': ['d2_m1'], 'expid_d2_m1_2_POST': ['d2_m1'], 'expid_d2_m2_1_POST': ['d2_m2'],
            'expid_d2_m2_2_POST': ['d2_m2'],

            'expid_d1_m1_1_CLEAN': ['d1_m1'], 'expid_d1_m1_2_CLEAN': ['d1_m1'], 'expid_d1_m2_1_CLEAN': ['d1_m2'],
            'expid_d1_m2_2_CLEAN': ['d1_m2'],
            'expid_d2_m1_1_CLEAN': ['d2_m1'], 'expid_d2_m1_2_CLEAN': ['d2_m1'], 'expid_d2_m2_1_CLEAN': ['d2_m2'],
            'expid_d2_m2_2_CLEAN': ['d2_m2'],

            'expid_d1_1_ASIM' : ['d1_m1', 'd1_m2'], 'expid_d1_2_ASIM' : ['d1_m1', 'd1_m2'],
            'expid_d2_1_ASIM': ['d2_m1', 'd2_m2'], 'expid_d2_2_ASIM' : ['d2_m1', 'd2_m2']
        }

        nodes = [
            "expid_SETUP",
            'd1_m1', 'd1_m2', 'd2_m2', 'd2_m1'
        ]
        edges = [
            ("expid_SETUP", "d1_m1"), ("expid_SETUP", "d1_m2"), ("expid_SETUP", "d2_m1"),
            ("expid_SETUP", "d2_m2"),

            ("d1_m1", "d1_m1"), ("d1_m2", "d1_m2"),
            ("d2_m1", "d2_m1"), ("d2_m2", "d2_m2")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict)

        self.assertTrue(graph.obj_dict['strict'])

        subgraphs = graph.obj_dict['subgraphs']
        experiment_subgraph = subgraphs['Experiment'][0]

        self.assertListEqual(sorted(list(experiment_subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(experiment_subgraph['edges'].keys())), sorted(edges))

        subgraph_synchronize_d1 = graph.obj_dict['subgraphs']['cluster_d1_m1_d1_m2'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_d1['nodes'].keys())), sorted(['d1_m1', 'd1_m2']))
        self.assertListEqual(sorted(list(subgraph_synchronize_d1['edges'].keys())), sorted([('d1_m1', 'd1_m2')]))

        subgraph_synchronize_d2 = graph.obj_dict['subgraphs']['cluster_d2_m1_d2_m2'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_d2['nodes'].keys())), sorted(['d2_m1', 'd2_m2']))
        self.assertListEqual(sorted(list(subgraph_synchronize_d2['edges'].keys())), sorted([('d2_m1', 'd2_m2')]))

    def test_synchronize_member_group_chunk(self):
        for date in ['d1', 'd2']:
            for chunk in [1, 2]:
                job = self._createDummyJob('expid_' + date + '_' + str(chunk) + '_ASIM',
                                           Status.WAITING, date, None, chunk)
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))
                self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'d1_m1_1': Status.WAITING, 'd1_m1_2': Status.WAITING,
                                 'd1_m2_1': Status.WAITING, 'd1_m2_2': Status.WAITING,
                                 'd2_m1_1': Status.WAITING, 'd2_m1_2': Status.WAITING,
                                 'd2_m2_1': Status.WAITING, 'd2_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_1_SIM': ['d1_m1_1'], 'expid_d1_m1_2_SIM': ['d1_m1_2'], 'expid_d1_m2_1_SIM': ['d1_m2_1'],
            'expid_d1_m2_2_SIM': ['d1_m2_2'],
            'expid_d2_m1_1_SIM': ['d2_m1_1'], 'expid_d2_m1_2_SIM': ['d2_m1_2'], 'expid_d2_m2_1_SIM': ['d2_m2_1'],
            'expid_d2_m2_2_SIM': ['d2_m2_2'],

            'expid_d1_m1_1_POST': ['d1_m1_1'], 'expid_d1_m1_2_POST': ['d1_m1_2'], 'expid_d1_m2_1_POST': ['d1_m2_1'],
            'expid_d1_m2_2_POST': ['d1_m2_2'],
            'expid_d2_m1_1_POST': ['d2_m1_1'], 'expid_d2_m1_2_POST': ['d2_m1_2'], 'expid_d2_m2_1_POST': ['d2_m2_1'],
            'expid_d2_m2_2_POST': ['d2_m2_2'],

            'expid_d1_m1_1_CLEAN': ['d1_m1_1'], 'expid_d1_m1_2_CLEAN': ['d1_m1_2'], 'expid_d1_m2_1_CLEAN': ['d1_m2_1'],
            'expid_d1_m2_2_CLEAN': ['d1_m2_2'],
            'expid_d2_m1_1_CLEAN': ['d2_m1_1'], 'expid_d2_m1_2_CLEAN': ['d2_m1_2'], 'expid_d2_m2_1_CLEAN': ['d2_m2_1'],
            'expid_d2_m2_2_CLEAN': ['d2_m2_2'],

            'expid_d1_1_ASIM' : ['d1_m1_1', 'd1_m2_1'], 'expid_d1_2_ASIM' : ['d1_m1_2', 'd1_m2_2'],
            'expid_d2_1_ASIM': ['d2_m1_1', 'd2_m2_1'], 'expid_d2_2_ASIM' : ['d2_m1_2', 'd2_m2_2']
        }

        nodes = [
            "expid_SETUP", "expid_d1_m1_INI", "expid_d1_m2_INI", "expid_d2_m1_INI", "expid_d2_m2_INI",
            'd1_m1_1', 'd1_m2_1', 'd2_m1_1', 'd2_m2_1', 'd1_m1_2', 'd1_m2_2', 'd2_m1_2', 'd2_m2_2'
        ]
        edges = [
            ("expid_SETUP", "expid_d1_m1_INI"), ("expid_SETUP", "expid_d1_m2_INI"), ("expid_SETUP", "expid_d2_m1_INI"),
            ("expid_SETUP", "expid_d2_m2_INI"), ("expid_d1_m1_INI", "d1_m1_1"), ("expid_d1_m2_INI", "d1_m2_1"),
            ("expid_d2_m1_INI", "d2_m1_1"), ("expid_d2_m2_INI", "d2_m2_1"),
            ("d1_m1_1", "d1_m1_2"), ("d1_m2_1", "d1_m2_2"), ("d2_m1_1", "d2_m1_2"), ("d2_m2_1", "d2_m2_2"),

            ("d1_m1_1", "d1_m1_1"), ("d1_m1_2", "d1_m1_2"), ("d1_m2_1", "d1_m2_1"), ("d1_m2_2", "d1_m2_2"),
            ("d2_m1_1", "d2_m1_1"), ("d2_m1_2", "d2_m1_2"), ("d2_m2_1", "d2_m2_1"), ("d2_m2_2", "d2_m2_2")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict)

        self.assertTrue(graph.obj_dict['strict'])

        subgraphs = graph.obj_dict['subgraphs']
        experiment_subgraph = subgraphs['Experiment'][0]

        self.assertListEqual(sorted(list(experiment_subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(experiment_subgraph['edges'].keys())), sorted(edges))

        subgraph_synchronize_d1_1 = graph.obj_dict['subgraphs']['cluster_d1_m1_1_d1_m2_1'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_d1_1['nodes'].keys())), sorted(['d1_m1_1', 'd1_m2_1']))
        self.assertListEqual(sorted(list(subgraph_synchronize_d1_1['edges'].keys())), sorted([('d1_m1_1', 'd1_m2_1')]))

        subgraph_synchronize_d1_2 = graph.obj_dict['subgraphs']['cluster_d1_m1_2_d1_m2_2'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_d1_2['nodes'].keys())), sorted(['d1_m1_2', 'd1_m2_2']))
        self.assertListEqual(sorted(list(subgraph_synchronize_d1_2['edges'].keys())), sorted([('d1_m1_2', 'd1_m2_2')]))

        subgraph_synchronize_d2_1 = graph.obj_dict['subgraphs']['cluster_d2_m1_1_d2_m2_1'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_d2_1['nodes'].keys())), sorted(['d2_m1_1', 'd2_m2_1']))
        self.assertListEqual(sorted(list(subgraph_synchronize_d2_1['edges'].keys())), sorted([('d2_m1_1', 'd2_m2_1')]))

        subgraph_synchronize_d2_2 = graph.obj_dict['subgraphs']['cluster_d2_m1_2_d2_m2_2'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_d2_2['nodes'].keys())), sorted(['d2_m1_2', 'd2_m2_2']))
        self.assertListEqual(sorted(list(subgraph_synchronize_d2_2['edges'].keys())), sorted([('d2_m1_2', 'd2_m2_2')]))

    def test_synchronize_member_group_date(self):
        for date in ['d1', 'd2']:
            for chunk in [1, 2]:
                job = self._createDummyJob('expid_' + date + '_' + str(chunk) + '_ASIM',
                                           Status.WAITING, date, None, chunk)
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

                self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'d1': Status.WAITING,
                                 'd2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_INI': ['d1'], 'expid_d1_m2_INI': ['d1'], 'expid_d2_m1_INI': ['d2'],
            'expid_d2_m2_INI': ['d2'],
            'expid_d1_m1_1_SIM': ['d1'], 'expid_d1_m1_2_SIM': ['d1'], 'expid_d1_m2_1_SIM': ['d1'],
            'expid_d1_m2_2_SIM': ['d1'],
            'expid_d2_m1_1_SIM': ['d2'], 'expid_d2_m1_2_SIM': ['d2'], 'expid_d2_m2_1_SIM': ['d2'],
            'expid_d2_m2_2_SIM': ['d2'],

            'expid_d1_m1_1_POST': ['d1'], 'expid_d1_m1_2_POST': ['d1'], 'expid_d1_m2_1_POST': ['d1'],
            'expid_d1_m2_2_POST': ['d1'],
            'expid_d2_m1_1_POST': ['d2'], 'expid_d2_m1_2_POST': ['d2'], 'expid_d2_m2_1_POST': ['d2'],
            'expid_d2_m2_2_POST': ['d2'],

            'expid_d1_m1_1_CLEAN': ['d1'], 'expid_d1_m1_2_CLEAN': ['d1'], 'expid_d1_m2_1_CLEAN': ['d1'],
            'expid_d1_m2_2_CLEAN': ['d1'],
            'expid_d2_m1_1_CLEAN': ['d2'], 'expid_d2_m1_2_CLEAN': ['d2'], 'expid_d2_m2_1_CLEAN': ['d2'],
            'expid_d2_m2_2_CLEAN': ['d2'],

            'expid_d1_1_ASIM': ['d1'], 'expid_d1_2_ASIM': ['d1'],
            'expid_d2_1_ASIM': ['d2'], 'expid_d2_2_ASIM': ['d2']
        }

        nodes = [
            "expid_SETUP",
            'd1', 'd2'
        ]
        edges = [
            ("expid_SETUP", "d1"), ("expid_SETUP", "d2"),

            ("d1", "d1"), ("d2", "d2")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict)

        self.assertTrue(graph.obj_dict['strict'])

        subgraphs = graph.obj_dict['subgraphs']
        experiment_subgraph = subgraphs['Experiment'][0]

        self.assertListEqual(sorted(list(experiment_subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(experiment_subgraph['edges'].keys())), sorted(edges))

        for subgraph in list(subgraphs.keys()):
            self.assertFalse(subgraph.startswith('cluster'))

    def test_synchronize_date_group_member(self):
        for chunk in [1, 2]:
            job = self._createDummyJob('expid_' + str(chunk) + '_ASIM',
                                           Status.WAITING, None, None, chunk)
            for date in ['d1', 'd2']:
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

            self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'d1_m1': Status.WAITING,
                                 'd1_m2': Status.WAITING,
                                 'd2_m1': Status.WAITING,
                                 'd2_m2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_INI': ['d1_m1'], 'expid_d1_m2_INI': ['d1_m2'], 'expid_d2_m1_INI': ['d2_m1'],
            'expid_d2_m2_INI': ['d2_m2'],
            'expid_d1_m1_1_SIM': ['d1_m1'], 'expid_d1_m1_2_SIM': ['d1_m1'], 'expid_d1_m2_1_SIM': ['d1_m2'],
            'expid_d1_m2_2_SIM': ['d1_m2'],
            'expid_d2_m1_1_SIM': ['d2_m1'], 'expid_d2_m1_2_SIM': ['d2_m1'], 'expid_d2_m2_1_SIM': ['d2_m2'],
            'expid_d2_m2_2_SIM': ['d2_m2'],

            'expid_d1_m1_1_POST': ['d1_m1'], 'expid_d1_m1_2_POST': ['d1_m1'], 'expid_d1_m2_1_POST': ['d1_m2'],
            'expid_d1_m2_2_POST': ['d1_m2'],
            'expid_d2_m1_1_POST': ['d2_m1'], 'expid_d2_m1_2_POST': ['d2_m1'], 'expid_d2_m2_1_POST': ['d2_m2'],
            'expid_d2_m2_2_POST': ['d2_m2'],

            'expid_d1_m1_1_CLEAN': ['d1_m1'], 'expid_d1_m1_2_CLEAN': ['d1_m1'], 'expid_d1_m2_1_CLEAN': ['d1_m2'],
            'expid_d1_m2_2_CLEAN': ['d1_m2'],
            'expid_d2_m1_1_CLEAN': ['d2_m1'], 'expid_d2_m1_2_CLEAN': ['d2_m1'], 'expid_d2_m2_1_CLEAN': ['d2_m2'],
            'expid_d2_m2_2_CLEAN': ['d2_m2'],

            'expid_1_ASIM' : ['d1_m1', 'd1_m2', 'd2_m1', 'd2_m2'], 'expid_2_ASIM' : ['d1_m1', 'd1_m2', 'd2_m1', 'd2_m2']
        }

        nodes = [
            "expid_SETUP",
            'd1_m1', 'd1_m2', 'd2_m2', 'd2_m1'
        ]
        edges = [
            ("expid_SETUP", "d1_m1"), ("expid_SETUP", "d1_m2"), ("expid_SETUP", "d2_m1"),
            ("expid_SETUP", "d2_m2"),

            ("d1_m1", "d1_m1"), ("d1_m2", "d1_m2"),
            ("d2_m1", "d2_m1"), ("d2_m2", "d2_m2")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict)

        self.assertTrue(graph.obj_dict['strict'])

        subgraphs = graph.obj_dict['subgraphs']
        experiment_subgraph = subgraphs['Experiment'][0]

        self.assertListEqual(sorted(list(experiment_subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(experiment_subgraph['edges'].keys())), sorted(edges))

        subgraph_synchronize_d1_d2 = graph.obj_dict['subgraphs']['cluster_d1_m1_d1_m2_d2_m1_d2_m2'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_d1_d2['nodes'].keys())), sorted(['d1_m1', 'd1_m2', 'd2_m1', 'd2_m2']))
        self.assertListEqual(sorted(list(subgraph_synchronize_d1_d2['edges'].keys())), sorted([('d1_m1', 'd1_m2'), ('d1_m2', 'd2_m1'), ('d2_m1', 'd2_m2')]))

    def test_synchronize_date_group_chunk(self):
        for chunk in [1, 2]:
            job = self._createDummyJob('expid_' + str(chunk) + '_ASIM',
                                           Status.WAITING, None, None, chunk)
            for date in ['d1', 'd2']:
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

            self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'d1_m1_1': Status.WAITING, 'd1_m1_2': Status.WAITING,
                                 'd1_m2_1': Status.WAITING, 'd1_m2_2': Status.WAITING,
                                 'd2_m1_1': Status.WAITING, 'd2_m1_2': Status.WAITING,
                                 'd2_m2_1': Status.WAITING, 'd2_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_1_SIM': ['d1_m1_1'], 'expid_d1_m1_2_SIM': ['d1_m1_2'], 'expid_d1_m2_1_SIM': ['d1_m2_1'],
            'expid_d1_m2_2_SIM': ['d1_m2_2'],
            'expid_d2_m1_1_SIM': ['d2_m1_1'], 'expid_d2_m1_2_SIM': ['d2_m1_2'], 'expid_d2_m2_1_SIM': ['d2_m2_1'],
            'expid_d2_m2_2_SIM': ['d2_m2_2'],

            'expid_d1_m1_1_POST': ['d1_m1_1'], 'expid_d1_m1_2_POST': ['d1_m1_2'], 'expid_d1_m2_1_POST': ['d1_m2_1'],
            'expid_d1_m2_2_POST': ['d1_m2_2'],
            'expid_d2_m1_1_POST': ['d2_m1_1'], 'expid_d2_m1_2_POST': ['d2_m1_2'], 'expid_d2_m2_1_POST': ['d2_m2_1'],
            'expid_d2_m2_2_POST': ['d2_m2_2'],

            'expid_d1_m1_1_CLEAN': ['d1_m1_1'], 'expid_d1_m1_2_CLEAN': ['d1_m1_2'], 'expid_d1_m2_1_CLEAN': ['d1_m2_1'],
            'expid_d1_m2_2_CLEAN': ['d1_m2_2'],
            'expid_d2_m1_1_CLEAN': ['d2_m1_1'], 'expid_d2_m1_2_CLEAN': ['d2_m1_2'], 'expid_d2_m2_1_CLEAN': ['d2_m2_1'],
            'expid_d2_m2_2_CLEAN': ['d2_m2_2'],

            'expid_1_ASIM' : ['d1_m1_1', 'd1_m2_1', 'd2_m1_1', 'd2_m2_1'], 'expid_2_ASIM' : ['d1_m1_2', 'd1_m2_2', 'd2_m1_2', 'd2_m2_2'],
        }

        nodes = [
            "expid_SETUP", "expid_d1_m1_INI", "expid_d1_m2_INI", "expid_d2_m1_INI", "expid_d2_m2_INI",
            'd1_m1_1', 'd1_m2_1', 'd2_m1_1', 'd2_m2_1', 'd1_m1_2', 'd1_m2_2', 'd2_m1_2', 'd2_m2_2'
        ]
        edges = [
            ("expid_SETUP", "expid_d1_m1_INI"), ("expid_SETUP", "expid_d1_m2_INI"), ("expid_SETUP", "expid_d2_m1_INI"),
            ("expid_SETUP", "expid_d2_m2_INI"), ("expid_d1_m1_INI", "d1_m1_1"), ("expid_d1_m2_INI", "d1_m2_1"),
            ("expid_d2_m1_INI", "d2_m1_1"), ("expid_d2_m2_INI", "d2_m2_1"),
            ("d1_m1_1", "d1_m1_2"), ("d1_m2_1", "d1_m2_2"), ("d2_m1_1", "d2_m1_2"), ("d2_m2_1", "d2_m2_2"),

            ("d1_m1_1", "d1_m1_1"), ("d1_m1_2", "d1_m1_2"), ("d1_m2_1", "d1_m2_1"), ("d1_m2_2", "d1_m2_2"),
            ("d2_m1_1", "d2_m1_1"), ("d2_m1_2", "d2_m1_2"), ("d2_m2_1", "d2_m2_1"), ("d2_m2_2", "d2_m2_2")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict)

        self.assertTrue(graph.obj_dict['strict'])

        subgraphs = graph.obj_dict['subgraphs']
        experiment_subgraph = subgraphs['Experiment'][0]

        self.assertListEqual(sorted(list(experiment_subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(experiment_subgraph['edges'].keys())), sorted(edges))

        subgraph_synchronize_1 = graph.obj_dict['subgraphs']['cluster_d1_m1_1_d1_m2_1_d2_m1_1_d2_m2_1'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_1['nodes'].keys())), sorted(['d1_m1_1', 'd1_m2_1', 'd2_m1_1', 'd2_m2_1']))
        self.assertListEqual(sorted(list(subgraph_synchronize_1['edges'].keys())), sorted([('d1_m1_1', 'd1_m2_1'), ('d1_m2_1', 'd2_m1_1'), ('d2_m1_1', 'd2_m2_1')]))

        subgraph_synchronize_2 = graph.obj_dict['subgraphs']['cluster_d1_m1_2_d1_m2_2_d2_m1_2_d2_m2_2'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_2['nodes'].keys())), sorted(['d1_m1_2', 'd1_m2_2', 'd2_m1_2', 'd2_m2_2']))
        self.assertListEqual(sorted(list(subgraph_synchronize_2['edges'].keys())), sorted([('d1_m1_2', 'd1_m2_2'), ('d1_m2_2', 'd2_m1_2'), ('d2_m1_2', 'd2_m2_2')]))

    def test_synchronize_date_group_date(self):
        for chunk in [1, 2]:
            job = self._createDummyJob('expid_' + str(chunk) + '_ASIM',
                                           Status.WAITING, None, None, chunk)
            for date in ['d1', 'd2']:
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

            self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'d1': Status.WAITING,
                                 'd2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_INI': ['d1'], 'expid_d1_m2_INI': ['d1'], 'expid_d2_m1_INI': ['d2'],
            'expid_d2_m2_INI': ['d2'],
            'expid_d1_m1_1_SIM': ['d1'], 'expid_d1_m1_2_SIM': ['d1'], 'expid_d1_m2_1_SIM': ['d1'],
            'expid_d1_m2_2_SIM': ['d1'],
            'expid_d2_m1_1_SIM': ['d2'], 'expid_d2_m1_2_SIM': ['d2'], 'expid_d2_m2_1_SIM': ['d2'],
            'expid_d2_m2_2_SIM': ['d2'],

            'expid_d1_m1_1_POST': ['d1'], 'expid_d1_m1_2_POST': ['d1'], 'expid_d1_m2_1_POST': ['d1'],
            'expid_d1_m2_2_POST': ['d1'],
            'expid_d2_m1_1_POST': ['d2'], 'expid_d2_m1_2_POST': ['d2'], 'expid_d2_m2_1_POST': ['d2'],
            'expid_d2_m2_2_POST': ['d2'],

            'expid_d1_m1_1_CLEAN': ['d1'], 'expid_d1_m1_2_CLEAN': ['d1'], 'expid_d1_m2_1_CLEAN': ['d1'],
            'expid_d1_m2_2_CLEAN': ['d1'],
            'expid_d2_m1_1_CLEAN': ['d2'], 'expid_d2_m1_2_CLEAN': ['d2'], 'expid_d2_m2_1_CLEAN': ['d2'],
            'expid_d2_m2_2_CLEAN': ['d2'],

            'expid_1_ASIM': ['d1', 'd2'], 'expid_2_ASIM': ['d1', 'd2']
        }

        nodes = [
            "expid_SETUP",
            'd1', 'd2'
        ]
        edges = [
            ("expid_SETUP", "d1"), ("expid_SETUP", "d2"),

            ("d1", "d1"), ("d2", "d2")
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, groups_dict)

        self.assertTrue(graph.obj_dict['strict'])

        subgraphs = graph.obj_dict['subgraphs']
        experiment_subgraph = subgraphs['Experiment'][0]

        self.assertListEqual(sorted(list(experiment_subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(experiment_subgraph['edges'].keys())), sorted(edges))

        subgraph_synchronize_d1_d2 = graph.obj_dict['subgraphs']['cluster_d1_d2'][0]
        self.assertListEqual(sorted(list(subgraph_synchronize_d1_d2['nodes'].keys())),
                             sorted(['d1', 'd2']))

    def test_normal_workflow(self):
        nodes = [
            "expid_SETUP", "expid_d1_m1_INI", "expid_d1_m2_INI", "expid_d2_m1_INI", "expid_d2_m2_INI",
            "expid_d1_m1_1_SIM", "expid_d1_m1_2_SIM", "expid_d1_m2_1_SIM", "expid_d1_m2_2_SIM",
            "expid_d2_m1_1_SIM", "expid_d2_m1_2_SIM", "expid_d2_m2_1_SIM", "expid_d2_m2_2_SIM",
            "expid_d1_m1_1_POST", "expid_d1_m1_2_POST", "expid_d1_m2_1_POST", "expid_d1_m2_2_POST",
            "expid_d2_m1_1_POST", "expid_d2_m1_2_POST", "expid_d2_m2_1_POST", "expid_d2_m2_2_POST",
            "expid_d1_m1_1_CLEAN", "expid_d1_m1_2_CLEAN", "expid_d1_m2_1_CLEAN", "expid_d1_m2_2_CLEAN",
            "expid_d2_m1_1_CLEAN", "expid_d2_m1_2_CLEAN", "expid_d2_m2_1_CLEAN", "expid_d2_m2_2_CLEAN"
        ]
        edges = [
            ('expid_SETUP', 'expid_d1_m1_INI'), ('expid_SETUP', 'expid_d1_m2_INI'), ('expid_SETUP', 'expid_d2_m1_INI'),
            ('expid_SETUP', 'expid_d2_m2_INI'),

            ('expid_d1_m1_INI', 'expid_d1_m1_1_SIM'), ('expid_d1_m2_INI', 'expid_d1_m2_1_SIM'),
            ('expid_d2_m1_INI', 'expid_d2_m1_1_SIM'), ('expid_d2_m2_INI', 'expid_d2_m2_1_SIM'),

            ('expid_d1_m1_1_SIM', 'expid_d1_m1_2_SIM'),
            ('expid_d1_m1_1_SIM', 'expid_d1_m1_1_POST'),

            ('expid_d1_m1_2_SIM', 'expid_d1_m1_2_POST'),

            ('expid_d1_m2_1_SIM', 'expid_d1_m2_2_SIM'),
            ('expid_d1_m2_1_SIM', 'expid_d1_m2_1_POST'),

            ('expid_d1_m2_2_SIM', 'expid_d1_m2_2_POST'),

            ('expid_d2_m1_1_SIM', 'expid_d2_m1_2_SIM'),
            ('expid_d2_m1_1_SIM', 'expid_d2_m1_1_POST'),

            ('expid_d2_m1_2_SIM', 'expid_d2_m1_2_POST'),

            ('expid_d2_m2_1_SIM', 'expid_d2_m2_2_SIM'),
            ('expid_d2_m2_1_SIM', 'expid_d2_m2_1_POST'),

            ('expid_d2_m2_2_SIM', 'expid_d2_m2_2_POST'),

            ('expid_d1_m1_1_POST', 'expid_d1_m1_1_CLEAN'),
            ('expid_d1_m1_2_POST', 'expid_d1_m1_2_CLEAN'),
            ('expid_d1_m2_1_POST', 'expid_d1_m2_1_CLEAN'),
            ('expid_d1_m2_2_POST', 'expid_d1_m2_2_CLEAN'),

            ('expid_d2_m1_1_POST', 'expid_d2_m1_1_CLEAN'),
            ('expid_d2_m1_2_POST', 'expid_d2_m1_2_CLEAN'),
            ('expid_d2_m2_1_POST', 'expid_d2_m2_1_CLEAN'),
            ('expid_d2_m2_2_POST', 'expid_d2_m2_2_CLEAN')
        ]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), None, dict())

        self.assertFalse(graph.obj_dict['strict'])

        subgraph = graph.obj_dict['subgraphs']['Experiment'][0]

        self.assertListEqual(sorted(list(subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(subgraph['edges'].keys())), sorted(edges))

    def test_wrapper_and_groups(self):
        groups_dict = dict()

        groups_dict['status'] = {'d1_m1_1': Status.FAILED, 'd1_m1_2': Status.READY, 'd1_m2_1': Status.RUNNING,
                                 'd2_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_d1_m1_1_SIM': ['d1_m1_1'], 'expid_d1_m1_2_SIM': ['d1_m1_2'], 'expid_d1_m2_1_SIM': ['d1_m2_1'],
            'expid_d2_m2_2_SIM': ['d2_m2_2'],

            'expid_d1_m1_1_POST': ['d1_m1_1'], 'expid_d1_m1_2_POST': ['d1_m1_2'], 'expid_d1_m2_1_POST': ['d1_m2_1'],
            'expid_d2_m2_2_POST': ['d2_m2_2'],

            'expid_d1_m1_1_CLEAN': ['d1_m1_1'], 'expid_d1_m1_2_CLEAN': ['d1_m1_2'], 'expid_d1_m2_1_CLEAN': ['d1_m2_1'],
            'expid_d2_m2_2_CLEAN': ['d2_m2_2']
        }

        nodes = [
            "expid_SETUP", "expid_d1_m1_INI", "expid_d1_m2_INI", "expid_d2_m1_INI", "expid_d2_m2_INI",
            'd1_m1_1', 'd1_m1_2', 'd1_m2_1', 'd2_m2_2',

            'expid_d1_m2_2_SIM', 'expid_d1_m2_2_POST', 'expid_d1_m2_2_CLEAN',
            'expid_d2_m1_1_SIM', 'expid_d2_m1_1_POST', 'expid_d2_m1_1_CLEAN',
            'expid_d2_m1_2_SIM', 'expid_d2_m1_2_POST', 'expid_d2_m1_2_CLEAN',
            'expid_d2_m2_1_SIM', 'expid_d2_m2_1_POST', 'expid_d2_m2_1_CLEAN'
        ]
        edges = [
            ("expid_SETUP", "expid_d1_m1_INI"), ("expid_SETUP", "expid_d1_m2_INI"), ("expid_SETUP", "expid_d2_m1_INI"),
            ("expid_SETUP", "expid_d2_m2_INI"),

            ("expid_d1_m1_INI", "d1_m1_1"), ("expid_d1_m2_INI", "d1_m2_1"), ("expid_d2_m1_INI", "expid_d2_m1_1_SIM"),
            ("expid_d2_m2_INI", "expid_d2_m2_1_SIM"),

            ("d1_m1_1", "d1_m1_2"), ("d1_m1_1", "d1_m1_1"), ("d1_m1_2", "d1_m1_2"), ("d1_m2_1", "d1_m2_1"), ("d2_m2_2", "d2_m2_2"),

            ("d1_m2_1", "expid_d1_m2_2_SIM"),
            ("expid_d1_m2_2_SIM", "expid_d1_m2_2_POST"),
            ("expid_d1_m2_2_POST", "expid_d1_m2_2_CLEAN"),

            ("expid_d2_m1_1_SIM", "expid_d2_m1_1_POST"),
            ("expid_d2_m1_1_POST", "expid_d2_m1_1_CLEAN"),

            ("expid_d2_m1_1_SIM", "expid_d2_m1_2_SIM"),
            ("expid_d2_m1_2_SIM", "expid_d2_m1_2_POST"),
            ("expid_d2_m1_2_POST", "expid_d2_m1_2_CLEAN"),

            ("expid_d2_m2_1_SIM", "expid_d2_m2_1_POST"),
            ("expid_d2_m2_1_POST", "expid_d2_m2_1_CLEAN"),
            ("expid_d2_m2_1_SIM", "d2_m2_2")
        ]

        packages = [('expid', 'package_d1_m1_SIM', 'expid_d1_m1_1_SIM'),
                    ('expid', 'package_d1_m1_SIM', 'expid_d1_m1_2_SIM'),
                    ('expid', 'package_d2_m2_SIM', 'expid_d2_m2_1_SIM'),
                    ('expid', 'package_d2_m2_SIM', 'expid_d2_m2_2_SIM')]

        monitor = Monitor()
        graph = monitor.create_tree_list(self.experiment_id, self.job_list.get_job_list(), packages, groups_dict)
        self.assertTrue(graph.obj_dict['strict'])

        for (expid, package, job_name) in packages:
            if package != 'package_d2_m2_SIM':
                self.assertNotIn('cluster_' + package, graph.obj_dict['subgraphs'])
            else:
                self.assertIn('cluster_' + package, graph.obj_dict['subgraphs'])

        subgraph = graph.obj_dict['subgraphs']['Experiment'][0]

        self.assertListEqual(sorted(list(subgraph['nodes'].keys())), sorted(nodes))
        self.assertListEqual(sorted(list(subgraph['edges'].keys())), sorted(edges))

    def _createDummyJob(self, name, status, date=None, member=None, chunk=None, split=None):
        job_id = randrange(1, 999)
        job = Job(name, job_id, status, 0)
        job.type = randrange(0, 2)

        job.date = date
        job.member = member
        job.chunk = chunk
        job.split = split

        return job
class FakeBasicConfig:
    def __init__(self):
        pass

    DB_DIR = '/dummy/db/dir'
    DB_FILE = '/dummy/db/file'
    DB_PATH = '/dummy/db/path'
    LOCAL_ROOT_DIR = '/dummy/local/root/dir'
    LOCAL_TMP_DIR = '/dummy/local/temp/dir'
    LOCAL_PROJ_DIR = '/dummy/local/proj/dir'
    DEFAULT_PLATFORMS_CONF = ''
    DEFAULT_JOBS_CONF = ''


from unittest import TestCase

from autosubmit.job.job_common import Status


class TestJobCommon(TestCase):
    """
        This test is intended to prevent wrong changes on the Status class definition
    """

    def test_value_to_key_has_the_same_values_as_status_constants(self):
        self.assertEqual('SUSPENDED', Status.VALUE_TO_KEY[Status.SUSPENDED])
        self.assertEqual('UNKNOWN', Status.VALUE_TO_KEY[Status.UNKNOWN])
        self.assertEqual('FAILED', Status.VALUE_TO_KEY[Status.FAILED])
        self.assertEqual('WAITING', Status.VALUE_TO_KEY[Status.WAITING])
        self.assertEqual('READY', Status.VALUE_TO_KEY[Status.READY])
        self.assertEqual('SUBMITTED', Status.VALUE_TO_KEY[Status.SUBMITTED])
        self.assertEqual('HELD', Status.VALUE_TO_KEY[Status.HELD])
        self.assertEqual('QUEUING', Status.VALUE_TO_KEY[Status.QUEUING])
        self.assertEqual('RUNNING', Status.VALUE_TO_KEY[Status.RUNNING])
        self.assertEqual('COMPLETED', Status.VALUE_TO_KEY[Status.COMPLETED])



#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import time
import random
import os
from shutil import copy2
from autosubmit.history.database_managers.experiment_history_db_manager import ExperimentHistoryDbManager
from autosubmit.history.database_managers.experiment_status_db_manager import ExperimentStatusDbManager
from autosubmit.history.data_classes.experiment_run import ExperimentRun
from autosubmit.history.data_classes.job_data import JobData
from autosubmitconfigparser.config.basicconfig import BasicConfig
import autosubmit.history.utils as HUtils
EXPID_TT00_SOURCE = "test_database.db~"
EXPID_TT01_SOURCE = "test_database_no_run.db~"
EXPID = "t024"
EXPID_NONE = "t027"
BasicConfig.read()
JOBDATA_DIR = BasicConfig.JOBDATA_DIR
LOCAL_ROOT_DIR = BasicConfig.LOCAL_ROOT_DIR

@unittest.skip('TODO: looks like another test that used actual experiments data')
class TestExperimentStatusDatabaseManager(unittest.TestCase):
  """ Covers Experiment Status Database Manager """
  def setUp(self):
    self.exp_status_db = ExperimentStatusDbManager(EXPID, BasicConfig.DB_DIR, BasicConfig.DB_FILE, LOCAL_ROOT_DIR)

  def test_get_current_experiment_status_row(self):    
    exp_status_row = self.exp_status_db.get_experiment_status_row_by_expid(EXPID)
    self.assertIsNotNone(exp_status_row)
    exp_status_row_none = self.exp_status_db.get_experiment_status_row_by_expid(EXPID_NONE)
    self.assertIsNone(exp_status_row_none)
    exp_row_direct = self.exp_status_db.get_experiment_status_row_by_exp_id(exp_status_row.exp_id)
    self.assertTrue(exp_status_row.exp_id == exp_row_direct.exp_id)


  def test_update_exp_status(self):    
    self.exp_status_db.update_exp_status(EXPID, "RUNNING")
    exp_status_row_current = self.exp_status_db.get_experiment_status_row_by_expid(EXPID)
    self.assertTrue(exp_status_row_current.status == "RUNNING")
    self.exp_status_db.update_exp_status(EXPID, "NOT RUNNING")
    exp_status_row_current = self.exp_status_db.get_experiment_status_row_by_expid(EXPID)
    self.assertTrue(exp_status_row_current.status == "NOT RUNNING")

  def test_create_exp_status(self):
    experiment = self.exp_status_db.get_experiment_row_by_expid(EXPID_NONE)
    self.exp_status_db.create_experiment_status_as_running(experiment)
    experiment_status = self.exp_status_db.get_experiment_status_row_by_expid(EXPID_NONE)
    self.assertIsNotNone(experiment_status)
    self.exp_status_db.delete_exp_status(EXPID_NONE)
    experiment_status = self.exp_status_db.get_experiment_status_row_by_expid(EXPID_NONE)
    self.assertIsNone(experiment_status)

@unittest.skip('TODO: looks like another test that used actual experiments data')
class TestExperimentHistoryDbManager(unittest.TestCase):
  """ Covers Experiment History Database Manager and Data Models """
  def setUp(self):
    self.experiment_database = ExperimentHistoryDbManager(EXPID, JOBDATA_DIR)    
    source_path_tt00 = os.path.join(JOBDATA_DIR, EXPID_TT00_SOURCE)
    self.target_path_tt00 = os.path.join(JOBDATA_DIR, "job_data_{0}.db".format(EXPID))        
    copy2(source_path_tt00, self.target_path_tt00)
    source_path_tt01 = os.path.join(JOBDATA_DIR, EXPID_TT01_SOURCE)
    self.target_path_tt01 = os.path.join(JOBDATA_DIR, "job_data_{0}.db".format(EXPID_NONE))
    copy2(source_path_tt01, self.target_path_tt01)
    self.experiment_database.initialize()    

  def tearDown(self):
    os.remove(self.target_path_tt00)
    os.remove(self.target_path_tt01)

  def test_get_max_id(self):        
    max_item = self.experiment_database.get_experiment_run_dc_with_max_id()  
    self.assertTrue(max_item.run_id > 0)
    self.assertTrue(max_item.run_id >= 18) # Max is 18

  def test_pragma(self):
    self.assertTrue(self.experiment_database._get_pragma_version() == 17) # Update version on changes

  def test_get_job_data(self):
    job_data = self.experiment_database._get_job_data_last_by_name("a29z_20000101_fc0_1_SIM")
    self.assertTrue(len(job_data) > 0)
    self.assertTrue(job_data[0].last == 1)
    self.assertTrue(job_data[0].job_name == "a29z_20000101_fc0_1_SIM")

    job_data = self.experiment_database.get_job_data_by_name("a29z_20000101_fc0_1_SIM")
    self.assertTrue(job_data[0].job_name == "a29z_20000101_fc0_1_SIM")

    job_data = self.experiment_database._get_job_data_last_by_run_id(18) # Latest
    self.assertTrue(len(job_data) > 0)

    job_data = self.experiment_database._get_job_data_last_by_run_id_and_finished(18) 
    self.assertTrue(len(job_data) > 0)

    job_data = self.experiment_database.get_job_data_all()
    self.assertTrue(len(job_data) > 0)

  def test_insert_and_delete_experiment_run(self):
    new_run = ExperimentRun(19)
    new_id = self.experiment_database._insert_experiment_run(new_run)
    self.assertIsNotNone(new_id)
    last_experiment_run = self.experiment_database.get_experiment_run_dc_with_max_id()
    self.assertTrue(new_id == last_experiment_run.run_id)
    self.experiment_database.delete_experiment_run(new_id)
    last_experiment_run = self.experiment_database.get_experiment_run_dc_with_max_id()
    self.assertTrue(new_id != last_experiment_run.run_id)

  def test_insert_and_delete_job_data(self):
    max_run_id = self.experiment_database.get_experiment_run_dc_with_max_id().run_id
    new_job_data_name = "test_001_name_{0}".format(int(time.time()))
    new_job_data = JobData(_id=1, job_name=new_job_data_name, run_id=max_run_id)
    new_job_data_id = self.experiment_database._insert_job_data(new_job_data)
    self.assertIsNotNone(new_job_data_id)
    self.experiment_database.delete_job_data(new_job_data_id)
    job_data = self.experiment_database.get_job_data_by_name(new_job_data_name)
    self.assertTrue(len(job_data) == 0)


  def test_update_experiment_run(self):
    experiment_run_data_class = self.experiment_database.get_experiment_run_dc_with_max_id() # 18
    backup_run = self.experiment_database.get_experiment_run_dc_with_max_id()
    experiment_run_data_class.chunk_unit = "unouno"
    experiment_run_data_class.running = random.randint(1, 100)
    experiment_run_data_class.queuing = random.randint(1, 100)
    experiment_run_data_class.suspended = random.randint(1, 100)
    self.experiment_database._update_experiment_run(experiment_run_data_class)
    last_experiment_run = self.experiment_database.get_experiment_run_dc_with_max_id() # 18
    self.assertTrue(last_experiment_run.chunk_unit == experiment_run_data_class.chunk_unit)
    self.assertTrue(last_experiment_run.running == experiment_run_data_class.running)
    self.assertTrue(last_experiment_run.queuing == experiment_run_data_class.queuing)
    self.assertTrue(last_experiment_run.suspended == experiment_run_data_class.suspended)
    self.experiment_database._update_experiment_run(backup_run)

  def test_job_data_from_model(self):
    job_data_rows = self.experiment_database._get_job_data_last_by_name("a29z_20000101_fc0_1_SIM")
    job_data_row_first = job_data_rows[0]
    job_data_data_class = JobData.from_model(job_data_row_first)    
    self.assertTrue(job_data_row_first.job_name == job_data_data_class.job_name)

  def test_update_job_data_processed(self):
    current_time = time.time()
    job_data_rows = self.experiment_database._get_job_data_last_by_name("a29z_20000101_fc0_1_SIM")
    job_data_row_first = job_data_rows[0]
    job_data_data_class = JobData.from_model(job_data_row_first)
    backup_job_dc = JobData.from_model(job_data_row_first)
    job_data_data_class.nnodes = random.randint(1, 1000)
    job_data_data_class.ncpus = random.randint(1, 1000)
    job_data_data_class.status = "DELAYED"
    job_data_data_class.finish = current_time
    self.experiment_database._update_job_data_by_id(job_data_data_class)
    job_data_rows_current = self.experiment_database._get_job_data_last_by_name("a29z_20000101_fc0_1_SIM")
    job_data_row_first = job_data_rows_current[0]
    self.assertTrue(job_data_row_first.nnodes == job_data_data_class.nnodes)
    self.assertTrue(job_data_row_first.ncpus == job_data_data_class.ncpus)
    self.assertTrue(job_data_row_first.status == job_data_data_class.status)
    self.assertTrue(job_data_row_first.finish == job_data_data_class.finish)
    self.experiment_database._update_job_data_by_id(backup_job_dc)

  def test_bulk_update(self):
    current_time = time.time()    
    all_job_data_rows = self.experiment_database.get_job_data_all()
    job_data_rows_test = [job for job in all_job_data_rows if job.run_id == 3]    
    backup = [JobData.from_model(job) for job in job_data_rows_test]
    list_job_data_class = [JobData.from_model(job) for job in job_data_rows_test]
    backup_changes = [(HUtils.get_current_datetime(), job.status, job.rowstatus, job._id) for job in list_job_data_class]
    changes = [(HUtils.get_current_datetime(), "DELAYED", job.rowstatus, job._id) for job in list_job_data_class]
    self.experiment_database.update_many_job_data_change_status(changes)
    all_job_data_rows = self.experiment_database.get_job_data_all()
    job_data_rows_validate = [job for job in all_job_data_rows if job.run_id == 3]
    for (job_val, change_item) in zip(job_data_rows_validate, changes):
      modified, status, rowstatus, _id = change_item
      # self.assertTrue(job_val.finish == finish)
      self.assertTrue(job_val.modified == modified)
      self.assertTrue(job_val.status == status)
      self.assertTrue(job_val.rowstatus == rowstatus)
      self.assertTrue(job_val.id == _id)
    self.experiment_database.update_many_job_data_change_status(backup_changes)

  def test_job_data_maxcounter(self):
    new_job_data = ExperimentHistoryDbManager(EXPID_NONE, JOBDATA_DIR) 
    new_job_data.initialize()   
    max_empty_table_counter = new_job_data.get_job_data_max_counter()    
    self.assertTrue(max_empty_table_counter == 0)
    max_existing_counter = self.experiment_database.get_job_data_max_counter()    
    self.assertTrue(max_existing_counter > 0)

  def test_register_submitted_job_data_dc(self):
    job_data_dc = self.experiment_database.get_job_data_dc_unique_latest_by_job_name("a29z_20000101_fc0_1_SIM")
    max_counter = self.experiment_database.get_job_data_max_counter()
    self.assertTrue(max_counter > 0)
    self.assertTrue(job_data_dc.counter > 0)
    next_counter = max(max_counter, job_data_dc.counter + 1)
    self.assertTrue(next_counter >= max_counter)
    self.assertTrue(next_counter >= job_data_dc.counter + 1)
    job_data_dc.counter = next_counter
    job_data_dc_current = self.experiment_database.register_submitted_job_data_dc(job_data_dc)
    self.assertTrue(job_data_dc._id < job_data_dc_current._id)
    job_data_last_list = self.experiment_database._get_job_data_last_by_name(job_data_dc.job_name)
    self.assertTrue(len(job_data_last_list) == 1)
    self.experiment_database.delete_job_data(job_data_last_list[0].id) 
    job_data_dc.last = 1
    updated_job_data_dc = self.experiment_database.update_job_data_dc_by_id(job_data_dc)
    self.assertTrue(job_data_dc._id == updated_job_data_dc._id)
    job_data_dc = self.experiment_database.get_job_data_dc_unique_latest_by_job_name("a29z_20000101_fc0_1_SIM")
    self.assertTrue(job_data_dc._id == updated_job_data_dc._id)
  
  def test_update_children_and_platform_output(self):
    job_data_dc = self.experiment_database.get_job_data_dc_unique_latest_by_job_name("a29z_20000101_fc0_1_SIM")
    children_str = "a00, a01, a02"
    platform_output_str = " SLURM OUTPUT "
    job_data_dc.children = children_str
    job_data_dc.platform_output = platform_output_str
    self.experiment_database.update_job_data_dc_by_id(job_data_dc)
    job_data_dc_updated = self.experiment_database.get_job_data_dc_unique_latest_by_job_name("a29z_20000101_fc0_1_SIM")
    self.assertTrue(job_data_dc_updated.children == children_str)
    self.assertTrue(job_data_dc_updated.platform_output == platform_output_str)
    # Back to normal
    job_data_dc.children = ""
    job_data_dc.platform_output = "NO OUTPUT"
    self.experiment_database.update_job_data_dc_by_id(job_data_dc)
    job_data_dc_updated = self.experiment_database.get_job_data_dc_unique_latest_by_job_name("a29z_20000101_fc0_1_SIM")
    self.assertTrue(job_data_dc_updated.children == "")
    self.assertTrue(job_data_dc_updated.platform_output == "NO OUTPUT")

  

  def test_experiment_run_dc(self):
    experiment_run = self.experiment_database.get_experiment_run_dc_with_max_id()
    self.assertIsNotNone(experiment_run)

  def test_if_database_exists(self):
    exp_manager = ExperimentHistoryDbManager("0000")
    self.assertTrue(exp_manager.my_database_exists() is False)


if __name__ == '__main__':
  unittest.main()


#!/usr/bin/python

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import os
from datetime import datetime

DATETIME_FORMAT = '%Y-%m-%d-%H:%M:%S'

def get_fields_as_comma_str(model):
  """ Get the fields of a namedtumple as a comma separated string. """
  return ",".join(model._fields)

def calculate_queue_time_in_seconds(submit_time, start_time):
  # type : (float, float) -> int
  """ Calculates queue time in seconds based on submit and start timestamps. """
  if submit_time > 0 and start_time > 0 and (start_time - submit_time) > 0:
    return int(start_time - submit_time)
  return 0

def calculate_run_time_in_seconds(start_time, finish_time):
  # type : (float, float) -> int  
  """ Calculates run time in seconds based on start and finish timestamps. """
  if finish_time > 0 and start_time > 0 and (finish_time - start_time) > 0:
    return int(finish_time - start_time)
  return 0

def get_current_datetime():
  # type : () -> str
  """ Returns the current time in format '%Y-%m-%d-%H:%M:%S' """
  return datetime.today().strftime(DATETIME_FORMAT)

def get_current_datetime_if_none(argument):
  # type : (Any) -> Union[Any, str]
  """ Returns the current time in format '%Y-%m-%d-%H:%M:%S' if the supplied argument is None, else return argument. """
  if argument is None:
    return get_current_datetime()
  else:
    return argument

def create_file_with_full_permissions(path):
  # type : (str) -> None
  """ creates a database files with full permissions """
  os.umask(0)
  os.open(path, os.O_WRONLY | os.O_CREAT, 0o777)

def create_path_if_not_exists(path):
  # type : (str) -> bool
  if not os.path.exists(path):
    os.makedirs(path)
    return True
  return False

def create_path_if_not_exists_group_permission(path):
  # type : (str) -> bool
  if not os.path.exists(path):
    os.umask(0)
    os.makedirs(path, mode=0o774)
    return True
  return False

class SupportedStatus:
  COMPLETED = "COMPLETED"
  FAILED = "FAILED"
  QUEUING = "QUEUING"
  SUBMITTED = "SUBMITTED"
  RUNNING = "RUNNING"
  SUSPENDED = "SUSPENDED"

# if __name__ == "__main__":
#   print(get_fields_as_comma_str())


#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABCMeta, abstractmethod

class PlatformMonitor(metaclass=ABCMeta):
  def __init__(self, platform_output):
    self.original_input = platform_output
    self.input = str(platform_output).strip()

  
  @abstractmethod
  def _identify_input_rows(self):
    """ """
  



#!/usr/bin/env python3

# Copyright 2017-2020 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.
import collections
from log.log import Log, AutosubmitCritical, AutosubmitError
from autosubmit.job.job_common import Status, Type
from bscearth.utils.date import sum_str_hours
from autosubmit.job.job_packages import JobPackageSimple, JobPackageVertical, JobPackageHorizontal, \
    JobPackageSimpleWrapped, JobPackageHorizontalVertical, JobPackageVerticalHorizontal, JobPackageBase
from operator import attrgetter
from math import ceil
import operator
from typing import List
import copy



class JobPackager(object):
    """
    Main class that manages Job wrapping.

    :param as_config: Autosubmit basic configuration.\n
    :type as_config: AutosubmitConfig object.\n
    :param platform: A particular platform we are dealing with, e.g. Lsf Platform.\n
    :type platform: Specific Platform Object, e.g. LsfPlatform(), EcPlatform(), ...\n
    :param jobs_list: Contains the list of the jobs, along other properties.\n
    :type jobs_list: JobList object.
    """

    def calculate_job_limits(self,platform,job=None):
        jobs_list = self._jobs_list
        # Submitted + Queuing Jobs for specific Platform
        queuing_jobs = jobs_list.get_queuing(platform)
        # We now consider the running jobs count
        running_jobs = jobs_list.get_running(platform)
        running_by_id = dict()
        for running_job in running_jobs:
            running_by_id[running_job.id] = running_job
        running_jobs_len = len(running_by_id.keys())

        queued_by_id = dict()
        for queued_job in queuing_jobs:
            queued_by_id[queued_job.id] = queued_job
        queuing_jobs_len = len(list(queued_by_id.keys()))

        submitted_jobs = jobs_list.get_submitted(platform)
        submitted_by_id = dict()
        for submitted_job in submitted_jobs:
            submitted_by_id[submitted_job.id] = submitted_job
        submitted_jobs_len = len(list(submitted_by_id.keys()))

        waiting_jobs = submitted_jobs_len + queuing_jobs_len
        # Calculate available space in Platform Queue
        if job is not None and job.max_waiting_jobs and platform.max_waiting_jobs and int(job.max_waiting_jobs) != int(platform.max_waiting_jobs):
            self._max_wait_jobs_to_submit = int(job.max_waiting_jobs) - int(waiting_jobs)
        else:
            self._max_wait_jobs_to_submit = int(platform.max_waiting_jobs) - int(waiting_jobs)
        # .total_jobs is defined in each section of platforms_.yml, if not from there, it comes form autosubmit_.yml
        # .total_jobs Maximum number of jobs at the same time
        if job is not None and job.total_jobs != platform.total_jobs:
            self._max_jobs_to_submit = job.total_jobs - queuing_jobs_len
        else:
            self._max_jobs_to_submit = platform.total_jobs - queuing_jobs_len
        # Subtracting running jobs
        self._max_jobs_to_submit = self._max_jobs_to_submit - running_jobs_len
        self._max_jobs_to_submit = self._max_jobs_to_submit if self._max_jobs_to_submit > 0 else 0
        self.max_jobs = min(self._max_wait_jobs_to_submit,self._max_jobs_to_submit)

    def __init__(self, as_config, platform, jobs_list, hold=False):
        self.current_wrapper_section = "WRAPPERS"
        self._as_config = as_config
        self._platform = platform
        self._jobs_list = jobs_list
        self._max_wait_jobs_to_submit = 9999999
        self.hold = hold
        # These are defined in the [wrapper] section of autosubmit_,conf
        self.wrapper_type = dict()
        self.wrapper_policy = dict()
        self.wrapper_method = dict()
        self.jobs_in_wrapper = dict()
        self.extensible_wallclock = dict()
        self.wrapper_info = list()
        self.calculate_job_limits(platform)
        self.special_variables = dict()


        #todo add default values
        #Wrapper building starts here
        for wrapper_section,wrapper_data in self._as_config.experiment_data.get("WRAPPERS",{}).items():
            if isinstance(wrapper_data,collections.abc.Mapping ):
                self.wrapper_type[wrapper_section] = self._as_config.get_wrapper_type(wrapper_data)
                self.wrapper_policy[wrapper_section] = self._as_config.get_wrapper_policy(wrapper_data)
                if self._as_config.get_wrapper_method(wrapper_data) is None:
                    self.wrapper_method[wrapper_section] = "asthread"
                else:
                    self.wrapper_method[wrapper_section] = self._as_config.get_wrapper_method(wrapper_data).lower()
                self.jobs_in_wrapper[wrapper_section] = self._as_config.get_wrapper_jobs(wrapper_data)
                self.extensible_wallclock[wrapper_section] = self._as_config.get_extensible_wallclock(wrapper_data)
        self.wrapper_info = [self.wrapper_type,self.wrapper_policy,self.wrapper_method,self.jobs_in_wrapper,self.extensible_wallclock] # to pass to job_packages
        Log.debug("Number of jobs available: {0}", self._max_wait_jobs_to_submit)
        if self.hold:
            Log.debug("Number of jobs prepared: {0}", len(jobs_list.get_prepared(platform)))
            if len(jobs_list.get_prepared(platform)) > 0:
                Log.debug("Jobs ready for {0}: {1}", self._platform.name, len(jobs_list.get_prepared(platform)))
        else:
            Log.debug("Number of jobs ready: {0}", len(jobs_list.get_ready(platform, hold=False)))
            if len(jobs_list.get_ready(platform)) > 0:
                Log.debug("Jobs ready for {0}: {1}", self._platform.name, len(jobs_list.get_ready(platform)))
        self._maxTotalProcessors = 0

    def compute_weight(self, job_list):
        job = self
        jobs_by_section = dict()
        held_jobs = self._jobs_list.get_held_jobs()
        jobs_held_by_section = dict()
        for job in held_jobs:
            if job.section not in jobs_held_by_section:
                jobs_held_by_section[job.section] = []
            jobs_held_by_section[job.section].append(job)
        for job in job_list:
            if job.section not in jobs_by_section:
                jobs_by_section[job.section] = []
            if job.status != Status.COMPLETED:
                jobs_by_section[job.section].append(job)

        for section in jobs_by_section:
            if section in list(jobs_held_by_section.keys()):
                weight = len(jobs_held_by_section[section]) + 1
            else:
                weight = 1
            highest_completed = []

            for job in sorted(jobs_by_section[section], key=operator.attrgetter('chunk')):
                weight = weight + 1
                job.distance_weight = weight
                completed_jobs = 9999
                if job.has_parents() > 1:
                    tmp = [
                        parent for parent in job.parents if parent.status == Status.COMPLETED]
                    if len(tmp) > completed_jobs:
                        completed_jobs = len(tmp)
                        highest_completed = [job]
                    else:
                        highest_completed.append(job)
            for job in highest_completed:
                job.distance_weight = job.distance_weight - 1
    def _special_variables(self,job):
        special_variables = dict()
        if job.section not in self.special_variables:
            special_variables[job.section] = dict()
            if job.total_jobs != self._platform.total_jobs:
                special_variables[job.section]["TOTAL_JOBS"] = job
                self.special_variables.update(special_variables)
    def build_packages(self):
        # type: () -> List[JobPackageBase]
        """
        Returns the list of the built packages to be submitted

        :return: List of packages depending on type of package, JobPackageVertical Object for 'vertical'.
        :rtype: List() of JobPackageVertical
        """
        packages_to_submit = list()
        # only_wrappers = False when coming from Autosubmit.submit_ready_jobs, jobs_filtered empty
        jobs_ready = list()
        if len(self._jobs_list.jobs_to_run_first) > 0:
            jobs_ready = [job for job in self._jobs_list.jobs_to_run_first if
                     ( self._platform is None or job.platform.name.upper() == self._platform.name.upper()) and
                     job.status == Status.READY]
        if len(jobs_ready) == 0:
            if self.hold:
                jobs_ready = self._jobs_list.get_prepared(self._platform)
            else:
                jobs_ready = self._jobs_list.get_ready(self._platform)

        if self.hold and len(jobs_ready) > 0:
            self.compute_weight(jobs_ready)
            sorted_jobs = sorted(
                jobs_ready, key=operator.attrgetter('distance_weight'))
            jobs_in_held_status = self._jobs_list.get_held_jobs() + self._jobs_list.get_submitted(self._platform, hold=self.hold)
            held_by_id = dict()
            for held_job in jobs_in_held_status:
                if held_job.id not in held_by_id:
                    held_by_id[held_job.id] = []
                held_by_id[held_job.id].append(held_job)
            current_held_jobs = len(list(held_by_id.keys()))
            remaining_held_slots = 5 - current_held_jobs
            Log.debug("there are currently {0} held jobs".format(remaining_held_slots))
            try:
                while len(sorted_jobs) > remaining_held_slots:
                    if sorted_jobs[-1].packed:
                        sorted_jobs[-1].packed = False
                    del sorted_jobs[-1]
                for job in sorted_jobs:
                    if job.distance_weight > 3:
                        sorted_jobs.remove(job)
                jobs_ready = sorted_jobs
                pass
            except IndexError:
                pass
        if len(jobs_ready) == 0:
            # If there are no jobs ready, result is tuple of empty
            return packages_to_submit
        #check if there are jobs listed on calculate_job_limits
        for job in jobs_ready:
            self._special_variables(job)
        if len(self.special_variables) > 0:
            for section in self.special_variables:
                if "TOTAL_JOBS" in self.special_variables[section]:
                    self.calculate_job_limits(self._platform,self.special_variables[section]["TOTAL_JOBS"])
                if not (self._max_wait_jobs_to_submit > 0 and self._max_jobs_to_submit > 0):
                    # If there is no more space in platform, result is tuple of empty
                    Log.debug("No more space in platform {0} for jobs {1}".format(self._platform.name,
                                                                                  [job.name for job in jobs_ready]))
                    return packages_to_submit
                self.calculate_job_limits(self._platform)

        else:
            self.calculate_job_limits(self._platform)
            if not (self._max_wait_jobs_to_submit > 0 and self._max_jobs_to_submit > 0):
                # If there is no more space in platform, result is tuple of empty
                Log.debug("No more space in platform {0} for jobs {1}".format(self._platform.name, [job.name for job in jobs_ready]))
                return packages_to_submit


        # Sort by 6 first digits of date
        available_sorted = sorted(
            jobs_ready, key=lambda k: k.long_name.split('_')[1][:6])
        # Sort by Priority, the highest first
        list_of_available = sorted(
            available_sorted, key=lambda k: k.priority, reverse=True)
        num_jobs_to_submit = min(self._max_wait_jobs_to_submit, len(jobs_ready), self._max_jobs_to_submit)
        # Take the first num_jobs_to_submit from the list of available
        jobs_to_submit_tmp = list_of_available[0:num_jobs_to_submit]
        #jobs_to_submit = [
        #    fresh_job for fresh_job in jobs_to_submit_tmp if fresh_job.fail_count == 0]
        jobs_to_submit = [fresh_job for fresh_job in jobs_to_submit_tmp]
        failed_wrapped_jobs = [failed_job for failed_job in jobs_to_submit_tmp if failed_job.fail_count > 0]
        for job in failed_wrapped_jobs:
            job.packed = False
        jobs_to_submit_by_section = self._divide_list_by_section(jobs_to_submit)
        # create wrapped package jobs Wrapper building starts here
        for wrapper_name,section_jobs in jobs_to_submit_by_section.items():
            self.current_wrapper_section = wrapper_name
            for section,jobs in section_jobs.items():
                if len(jobs) > 0:
                    if self.current_wrapper_section != "SIMPLE" and not self._platform.allow_wrappers:
                        Log.warning("Platform {0} does not allow wrappers, submitting jobs individually".format(self._platform.name))
                    if  wrapper_name != "SIMPLE" and self._platform.allow_wrappers and self.wrapper_type[self.current_wrapper_section] in ['horizontal', 'vertical','vertical-horizontal', 'horizontal-vertical'] :
                        # Trying to find the value in jobs_parser, if not, default to an autosubmit_.yml value (Looks first in [wrapper] section)
                        wrapper_limits = dict()
                        wrapper_limits["max_by_section"] = dict()
                        wrapper_limits["max"] = int(self._as_config.get_max_wrapped_jobs(self._as_config.experiment_data["WRAPPERS"][self.current_wrapper_section]))
                        wrapper_limits["max_v"] = int(self._as_config.get_max_wrapped_jobs_vertical(self._as_config.experiment_data["WRAPPERS"][self.current_wrapper_section]))
                        wrapper_limits["max_h"] = int(self._as_config.get_max_wrapped_jobs_horizontal(self._as_config.experiment_data["WRAPPERS"][self.current_wrapper_section]))
                        if wrapper_limits["max"] < wrapper_limits["max_v"] * wrapper_limits["max_h"]:
                            wrapper_limits["max"] = wrapper_limits["max_v"] * wrapper_limits["max_h"]
                        if wrapper_limits["max_v"] == -1:
                            wrapper_limits["max_v"] = wrapper_limits["max"]
                        if wrapper_limits["max_h"] == -1:
                            wrapper_limits["max_h"] = wrapper_limits["max"]
                        if '&' not in section:
                            dependencies_keys = self._as_config.jobs_data[section].get('DEPENDENCIES', "")
                            wrapper_limits["max_by_section"][section] = wrapper_limits["max"]
                            wrapper_limits["min"] = min(self._as_config.jobs_data[section].get(
                                "MIN_WRAPPED", 99999999), 0)
                        else:
                            multiple_sections = section.split('&')
                            dependencies_keys = []
                            min_value = int(self._as_config.get_min_wrapped_jobs(self._as_config.experiment_data["WRAPPERS"][self.current_wrapper_section]))
                            for sectionN in multiple_sections:
                                if self._as_config.jobs_data[sectionN].get('DEPENDENCIES',"") != "":
                                    dependencies_keys += self._as_config.jobs_data.get("DEPENDENCIES", "").split()
                                if self._as_config.jobs_data[sectionN].get('MAX_WRAPPED',None) is not None and len(str(self._as_config.jobs_data[sectionN].get('MAX_WRAPPED',None))) > 0:
                                    wrapper_limits["max_by_section"][sectionN] = int(self._as_config.jobs_data[sectionN].get("MAX_WRAPPED"))
                                else:
                                    wrapper_limits["max_by_section"][sectionN] = wrapper_limits["max"]
                                wrapper_limits["min"] = min(self._as_config.jobs_data[sectionN].get("MIN_WRAPPED",min_value),min_value)
                        hard_limit_wrapper =  wrapper_limits["max"]
                        wrapper_limits["min"] = min(wrapper_limits["min"], hard_limit_wrapper)
                        wrapper_limits["min_v"] = self._as_config.get_min_wrapped_jobs_vertical(self._as_config.experiment_data["WRAPPERS"][self.current_wrapper_section])
                        wrapper_limits["min_h"] = self._as_config.get_min_wrapped_jobs_horizontal(self._as_config.experiment_data["WRAPPERS"][self.current_wrapper_section])
                        wrapper_limits["max"] = hard_limit_wrapper
                        if wrapper_limits["min"] < wrapper_limits["min_v"] * wrapper_limits["min_h"]:
                            wrapper_limits["min"] = max(wrapper_limits["min_v"],wrapper_limits["min_h"])
                        if len(self._jobs_list.jobs_to_run_first) > 0:
                            wrapper_limits["min"] = 2
                        current_info = list()
                        for param in self.wrapper_info:
                            current_info.append(param[self.current_wrapper_section])
                        if self.wrapper_type[self.current_wrapper_section] == 'vertical':
                            built_packages_tmp = self._build_vertical_packages(jobs, wrapper_limits,wrapper_info=current_info)
                        elif self.wrapper_type[self.current_wrapper_section] == 'horizontal':
                            built_packages_tmp = self._build_horizontal_packages(jobs, wrapper_limits, section,wrapper_info=current_info)
                        elif self.wrapper_type[self.current_wrapper_section] in ['vertical-horizontal', 'horizontal-vertical']:
                            built_packages_tmp = list()
                            built_packages_tmp.append(self._build_hybrid_package(jobs, wrapper_limits, section,wrapper_info=current_info))
                        else:
                            built_packages_tmp = self._build_vertical_packages(jobs, wrapper_limits)

                        for p in built_packages_tmp:
                            infinite_deadlock = False  # This will raise an autosubmit critical if true
                            failed_innerjobs = False
                            job_has_to_run_first = False
                            aux_jobs = []
                            # Check failed jobs first
                            for job in p.jobs:
                                job.wrapper_type = p.wrapper_type
                                if len(self._jobs_list.jobs_to_run_first) > 0:
                                    if job not in self._jobs_list.jobs_to_run_first:
                                        job.packed = False
                                        aux_jobs.append(job)
                                if job.fail_count > 0:
                                    failed_innerjobs = True
                            if len(self._jobs_list.jobs_to_run_first) > 0:
                                job_has_to_run_first = True
                                for job in aux_jobs:
                                    job.packed = False
                                    p.jobs.remove(job)
                                    if self.wrapper_type[self.current_wrapper_section] != "horizontal" and self.wrapper_type[self.current_wrapper_section] != "vertical" and self.wrapper_type[self.current_wrapper_section] != "vertical-mixed":
                                        for seq in range(0,len(p.jobs_lists)):
                                            try:
                                                p.jobs_lists[seq].remove(job)
                                            except Exception as e:
                                                pass
                                if self.wrapper_type[self.current_wrapper_section] != "horizontal" and self.wrapper_type[self.current_wrapper_section] != "vertical" and self.wrapper_type[self.current_wrapper_section] != "vertical-mixed":
                                    aux = p.jobs_lists
                                    p.jobs_lists = []
                                    for seq in range(0,len(aux)):
                                        if len(aux[seq]) > 0:
                                            p.jobs_lists.append(aux[seq])
                            if len(p.jobs) > 0:
                                balanced = True
                                if self.wrapper_type[self.current_wrapper_section] == 'vertical-horizontal':
                                    min_h = len(p.jobs_lists)
                                    min_v = len(p.jobs_lists[0])
                                    for list_of_jobs in p.jobs_lists[1:-1]:
                                        min_v = min(min_v, len(list_of_jobs))

                                elif self.wrapper_type[self.current_wrapper_section] == 'horizontal-vertical':
                                    min_v = len(p.jobs_lists)
                                    min_h = len(p.jobs_lists[0])
                                    i = 0
                                    for list_of_jobs in p.jobs_lists[1:-1]:
                                        min_h = min(min_h, len(list_of_jobs))
                                    for list_of_jobs in p.jobs_lists[:]:
                                        i = i+1
                                        if min_h != len(list_of_jobs) and i < len(p.jobs_lists):
                                            balanced = False
                                        elif min_h != len(list_of_jobs) and i == len(p.jobs_lists):
                                            if balanced:
                                                for job in list_of_jobs:
                                                    job.packed = False
                                                    p.jobs.remove(job)
                                                    package = JobPackageSimple([job])
                                                    packages_to_submit.append(package)
                                                p.jobs_lists = p.jobs_lists[:-1]



                                elif self.wrapper_type[self.current_wrapper_section] == 'horizontal':
                                    min_h = len(p.jobs)
                                    min_v = 1
                                elif self.wrapper_type[self.current_wrapper_section] == 'vertical':
                                    min_v = len(p.jobs)
                                    min_h = 1
                                else:
                                    min_v = len(p.jobs)
                                    min_h = len(p.jobs)
                                # if the quantity is enough, make the wrapper

                                if (len(p.jobs) >= wrapper_limits["min"] and min_v >= wrapper_limits["min_v"] and min_h >= wrapper_limits["min_h"] and (not failed_innerjobs or self.wrapper_policy[self.current_wrapper_section] not in ["mixed","strict"] ) ) or job_has_to_run_first:
                                    for job in p.jobs:
                                        job.packed = True
                                    packages_to_submit.append(p)
                                else:
                                    deadlock = True
                                    if deadlock: # Remaining jobs if chunk is the last one
                                        for job in p.jobs:
                                            if ( job.running == "chunk" and job.chunk == int(job.parameters["EXPERIMENT.NUMCHUNKS"]) ) and  balanced:
                                                deadlock = False
                                                break
                                    if not deadlock: # Submit package if deadlock has been liberated
                                        for job in p.jobs:
                                            job.packed = True
                                        packages_to_submit.append(p)
                                    else:
                                        wallclock_sum = p.jobs[0].wallclock
                                        for seq in range(1, min_v):
                                            wallclock_sum = sum_str_hours(wallclock_sum, p.jobs[0].wallclock)
                                        next_wrappable_jobs = self._jobs_list.get_jobs_by_section(self.jobs_in_wrapper[self.current_wrapper_section])
                                        next_wrappable_jobs = [job for job in next_wrappable_jobs if job.status == Status.WAITING and job not in p.jobs ] # Get only waiting jobs
                                        active_jobs = list()
                                        aux_active_jobs = list()
                                        for job in next_wrappable_jobs: # Prone tree by looking only the closest children
                                            direct_children = False
                                            for related in job.parents:
                                                if related in p.jobs:
                                                    direct_children = True
                                                    break
                                            if direct_children: # Get parent of direct children that aren't in wrapper
                                                aux_active_jobs += [aux_parent for aux_parent in job.parents if (  aux_parent.status != Status.COMPLETED and aux_parent.status != Status.FAILED) and ( aux_parent.section not in self.jobs_in_wrapper[self.current_wrapper_section] or ( aux_parent.section in self.jobs_in_wrapper[self.current_wrapper_section] and aux_parent.status != Status.COMPLETED and aux_parent.status != Status.FAILED and aux_parent.status != Status.WAITING and aux_parent.status != Status.READY ) ) ]
                                        aux_active_jobs = list(set(aux_active_jobs))
                                        track = [] # Tracker to prone tree for avoid the checking of the same parent from different nodes.
                                        active_jobs_names = [ job.name for job in p.jobs ] # We want to search if the actual wrapped jobs needs to run for add more jobs to this wrapper
                                        hard_deadlock = False
                                        for job in aux_active_jobs:
                                            parents_to_check = []
                                            if job.status == Status.WAITING: # We only want to check uncompleted parents
                                                aux_job = job
                                                for parent in aux_job.parents: # First case
                                                    if parent.name in active_jobs_names:
                                                        hard_deadlock = True
                                                        infinite_deadlock = True
                                                        break
                                                    if (parent.status == Status.WAITING ) and parent.name != aux_job.name:
                                                        parents_to_check.append(parent)
                                                track.extend(parents_to_check)
                                                while len(parents_to_check) > 0 and not infinite_deadlock: # We want to look deeper on the tree until all jobs are completed, or we find an unresolvable deadlock.
                                                    aux_job = parents_to_check.pop(0)
                                                    for parent in aux_job.parents:
                                                        if parent.name in active_jobs_names:
                                                            hard_deadlock = True
                                                            infinite_deadlock = True
                                                            break
                                                        if (parent.status == Status.WAITING ) and parent.name != aux_job.name and parent not in track:
                                                            parents_to_check.append(parent)
                                                    track.extend(parents_to_check)
                                            if not infinite_deadlock:
                                                active_jobs.append(job)  # List of jobs that can continue to run without run this wrapper
                                        # Act in base of active_jobs and Policies
                                        if self.wrapper_policy[self.current_wrapper_section] == "strict":
                                            error = True
                                            for job in p.jobs:
                                                job.packed = False
                                                if job in self._jobs_list.jobs_to_run_first:
                                                    error = False
                                                    if job.status == Status.READY:
                                                        if job.type == Type.PYTHON and not self._platform.allow_python_jobs:
                                                            package = JobPackageSimpleWrapped(
                                                                [job])
                                                        else:
                                                            package = JobPackageSimple([job])
                                                        packages_to_submit.append(package)
                                            if error:
                                                if len(active_jobs) > 0:
                                                    Log.printlog(
                                                        "Wrapper policy is set to MIXED and there are not enough jobs to form a wrapper.[wrappable:{4} <= defined_min:{5}] [wrappeable_h:{0} <= defined_min_h:{1}]|[wrappeable_v:{2} <= defined_min_v:{3}] waiting until the wrapper can be formed.\nIf all values are <=, some innerjob has failed under strict policy".format(
                                                            min_h, wrapper_limits["min_h"], min_v,
                                                            wrapper_limits["min_v"], wrapper_limits["min"], len(active_jobs)),
                                                        6013)
                                                else:
                                                    message = "Wrapper couldn't be formed under {0} POLICY due minimum limit not being reached: [wrappable:{4} < defined_min:{5}] [wrappable_h:{1} < defined_min_h:{2}]|[wrappeable_v:{3} < defined_min_v:{4}] ".format(
                                                        self.wrapper_policy[self.current_wrapper_section], min_h,
                                                        wrapper_limits["min_h"], min_v, wrapper_limits["min_v"],
                                                        wrapper_limits["min"], len(active_jobs))
                                                    if hard_deadlock:
                                                        message += "\nCheck your configuration: The next wrappable job can't be wrapped until some of inner jobs of current packages finishes which is imposible"
                                                    if min_v > 1:
                                                        message += "\nCheck your configuration: Check if current {0} vertical wallclock has reached the max defined on platforms.conf.".format(wallclock_sum)
                                                    else:
                                                        message += "\nCheck your configuration: Only jobs_in_wrappers are active, check their dependencies."
                                                    if not balanced:
                                                        message += "\nPackages are not well balanced: Check your dependencies(This is not the main cause of the Critical error)"
                                                    if len(self._jobs_list.get_in_queue()) == 0:
                                                        raise AutosubmitCritical(message, 7014)
                                        elif self.wrapper_policy[self.current_wrapper_section] == "mixed":
                                            error = True
                                            show_log = True
                                            for job in p.jobs:
                                                if job in self._jobs_list.jobs_to_run_first:
                                                    job.packed = False
                                                    error = False
                                                    if job.status == Status.READY:
                                                        if job.type == Type.PYTHON and not self._platform.allow_python_jobs:
                                                            package = JobPackageSimpleWrapped(
                                                                [job])
                                                        else:
                                                            package = JobPackageSimple([job])
                                                        packages_to_submit.append(package)
                                                if job.fail_count > 0 and job.status == Status.READY:
                                                    job.packed = False
                                                    Log.printlog(
                                                        "Wrapper policy is set to mixed, there is a failed job that will be sent sequential")
                                                    error = False
                                                    show_log = False
                                                    if job.type == Type.PYTHON and not self._platform.allow_python_jobs:
                                                        package = JobPackageSimpleWrapped(
                                                            [job])
                                                    else:
                                                        package = JobPackageSimple([job])
                                                    packages_to_submit.append(package)
                                            if error:
                                                if len(active_jobs) > 0:
                                                    if show_log:
                                                        Log.printlog(
                                                            "Wrapper policy is set to MIXED and there are not enough jobs to form a wrapper.[wrappable:{4} < defined_min:{5}] [wrappable_h:{0} < defined_min_h:{1}]|[wrappeable_v:{2} < defined_min_v:{3}] waiting until the wrapper can be formed.".format(
                                                                min_h, wrapper_limits["min_h"], min_v,
                                                                wrapper_limits["min_v"],wrapper_limits["min"],len(active_jobs)), 6013)
                                                else:
                                                    message = "Wrapper couldn't be formed under {0} POLICY due minimum limit not being reached: [wrappable:{4} < defined_min:{5}] [wrappable_h:{1} < defined_min_h:{2}]|[wrappeable_v:{3} < defined_min_v:{4}] ".format(
                                                            self.wrapper_policy[self.current_wrapper_section], min_h,
                                                            wrapper_limits["min_h"], min_v, wrapper_limits["min_v"],wrapper_limits["min"],len(active_jobs))
                                                    if hard_deadlock:
                                                        message += "\nCheck your configuration: The next wrappable job can't be wrapped until some of inner jobs of current packages finishes which is impossible"
                                                    if min_v > 1:
                                                        message += "\nCheck your configuration: Check if current {0} vertical wallclock has reached the max defined on platforms.conf.".format(
                                                            wallclock_sum)
                                                    else:
                                                        message += "\nCheck your configuration: Only jobs_in_wrappers are active, check your jobs_in_wrapper dependencies."
                                                    if not balanced:
                                                        message += "\nPackages are not well balanced! (This is not the main cause of the Critical error)"

                                                    if len(self._jobs_list.get_in_queue()) == 0: # When there are not more possible jobs, autosubmit will stop the execution
                                                        raise AutosubmitCritical(message, 7014)
                                        else:
                                            for job in p.jobs:
                                                job.packed = False
                                                if job.status == Status.READY:
                                                    if job.type == Type.PYTHON and not self._platform.allow_python_jobs:
                                                        package = JobPackageSimpleWrapped(
                                                            [job])
                                                    else:
                                                        package = JobPackageSimple([job])
                                                    packages_to_submit.append(package)
                                            Log.info("Wrapper policy is set to flexible and there is a deadlock, Autosubmit will submit the jobs sequentially")
                    else:
                        for job in jobs:
                            job.packed = False
                            if job.type == Type.PYTHON and not self._platform.allow_python_jobs:
                                package = JobPackageSimpleWrapped([job])
                            else:
                                package = JobPackageSimple([job])
                            packages_to_submit.append(package)


        for package in packages_to_submit:
            self.max_jobs = self.max_jobs - 1
            package.hold = self.hold

        return packages_to_submit

    def _divide_list_by_section(self, jobs_list):
        """
        Returns a dict() with as many keys as 'jobs_list' different sections
        The value for each key is a list() with all the jobs with the key section.

        :param jobs_list: list of jobs to be divided
        :rtype: Dictionary Key: Section Name, Value: List(Job Object)
        """
        # .jobs_in_wrapper defined in .yml, see constructor.
        sections_split = dict()
        jobs_by_section = dict()

        for wrapper_name,jobs_in_wrapper in self.jobs_in_wrapper.items():
            section_name = ""
            for section in jobs_in_wrapper:
                section_name += section+"&"
            section_name = section_name[:-1]
            sections_split[wrapper_name] = section_name
            jobs_by_section[wrapper_name] = dict()
            jobs_by_section[wrapper_name][section_name] = list()

        jobs_by_section["SIMPLE"] = collections.defaultdict(list)
        remaining_jobs = copy.copy(jobs_list)
        for wrapper_name,section_name in sections_split.items():
            for job in jobs_list:
                if job.section.upper() in section_name.split("&"):
                    jobs_by_section[wrapper_name][section_name].append(job)
                    try:
                        remaining_jobs.remove(job)
                    except ValueError:
                        pass
        for job in remaining_jobs:
            jobs_by_section["SIMPLE"][job.section].append(job)
        return jobs_by_section


    def _build_horizontal_packages(self, section_list, wrapper_limits, section,wrapper_info={}):
        packages = []
        horizontal_packager = JobPackagerHorizontal(section_list, self._platform.max_processors, wrapper_limits,
                                                    wrapper_limits["max"], self._platform.processors_per_node, self.wrapper_method[self.current_wrapper_section])

        package_jobs = horizontal_packager.build_horizontal_package()

        jobs_resources = dict()

        current_package = None
        if package_jobs:
            machinefile_function = self._as_config.get_wrapper_machinefiles()
            if machinefile_function == 'COMPONENTS':
                jobs_resources = horizontal_packager.components_dict
            jobs_resources['MACHINEFILES'] = machinefile_function
            current_package = JobPackageHorizontal(
                package_jobs, jobs_resources=jobs_resources, method=self.wrapper_method[self.current_wrapper_section], configuration=self._as_config, wrapper_section=self.current_wrapper_section)
            packages.append(current_package)

        return packages

    def _build_vertical_packages(self, section_list, wrapper_limits,wrapper_info={}):
        """
        Builds Vertical-Mixed or Vertical

        :param section_list: Jobs defined as wrappable belonging to a common section.\n
        :type section_list: List() of Job Objects. \n
        :param wrapper_limits: All wrapper limitations are inside this dictionary ( min,max,by_section,horizontal and vertical). \n
        :type wrapper_limits: Dict. \n
        :param wrapper_section: Current Section
        :type string
        :return: List of Wrapper Packages, Dictionary that details dependencies. \n
        :rtype: List() of JobPackageVertical(), Dictionary Key: String, Value: (Dictionary Key: Variable Name, Value: String/Int)
        """
        packages = []
        for job in section_list:
            if wrapper_limits["max"] > 0:
                if job.packed is False:
                    job.packed = True
                    dict_jobs = self._jobs_list.get_ordered_jobs_by_date_member(self.current_wrapper_section)
                    job_vertical_packager = JobPackagerVerticalMixed(dict_jobs, job, [job], job.wallclock, wrapper_limits["max"], wrapper_limits, self._platform.max_wallclock)
                    jobs_list = job_vertical_packager.build_vertical_package(job)

                    packages.append(JobPackageVertical(jobs_list, configuration=self._as_config,wrapper_section=self.current_wrapper_section,wrapper_info=wrapper_info))

            else:
                break
        return packages

    def _build_hybrid_package(self, jobs_list, wrapper_limits, section,wrapper_info={}):
        jobs_resources = dict()
        jobs_resources['MACHINEFILES'] = self._as_config.get_wrapper_machinefiles()

        ## READY JOBS ##
        ## Create the horizontal ##
        horizontal_packager = JobPackagerHorizontal(jobs_list, self._platform.max_processors, wrapper_limits,
                                                    wrapper_limits["max"], self._platform.processors_per_node,self.wrapper_method[self.current_wrapper_section])

        if self.wrapper_type[self.current_wrapper_section] == 'vertical-horizontal':
            return self._build_vertical_horizontal_package(horizontal_packager, jobs_resources)
        else:
            return self._build_horizontal_vertical_package(horizontal_packager, section, jobs_resources)

    def _build_horizontal_vertical_package(self, horizontal_packager, section, jobs_resources):
        total_wallclock = '00:00'
        horizontal_package = horizontal_packager.build_horizontal_package()
        horizontal_packager.create_sections_order(section)
        horizontal_packager.add_sectioncombo_processors(
            horizontal_packager.total_processors)
        horizontal_package.sort(
            key=lambda job: horizontal_packager.sort_by_expression(job.name))
        job = max(horizontal_package, key=attrgetter('total_wallclock'))
        wallclock = job.wallclock
        current_package = [horizontal_package]
        #current_package = []
        ## Get the next horizontal packages ##
        max_procs = horizontal_packager.total_processors
        new_package = horizontal_packager.get_next_packages(
            section, max_wallclock=self._platform.max_wallclock, horizontal_vertical=True, max_procs=max_procs)

        if new_package is not None and len(str(new_package)) > 0:
            current_package += new_package

        for i in range(len(current_package)):
            total_wallclock = sum_str_hours(total_wallclock, wallclock)
        if len(current_package) > 1:
            for level in range(1, len(current_package)):
                for job in current_package[level]:
                    job.level = level
        return JobPackageHorizontalVertical(current_package, max_procs, total_wallclock,
                                            jobs_resources=jobs_resources, configuration=self._as_config, wrapper_section=self.current_wrapper_section)

    def _build_vertical_horizontal_package(self, horizontal_packager, jobs_resources):
        total_wallclock = '00:00'
        horizontal_package = horizontal_packager.build_horizontal_package()
        total_processors = horizontal_packager.total_processors
        current_package = []
        ## Create the vertical ##
        actual_wrapped_jobs = len(horizontal_package)
        for job in horizontal_package:
            for section in horizontal_packager.wrapper_limits["max_by_section"]:
                if job.section == section:
                    horizontal_packager.wrapper_limits["max_by_section"][section] = horizontal_packager.wrapper_limits["max_by_section"][section] - 1
        horizontal_packager.wrapper_limits["max"] = horizontal_packager.wrapper_limits["max"] - actual_wrapped_jobs
        for job in horizontal_package:
            dict_jobs = self._jobs_list.get_ordered_jobs_by_date_member(self.current_wrapper_section)
            job_list = JobPackagerVerticalMixed(dict_jobs, job, [job], job.wallclock,
                                                             horizontal_packager.wrapper_limits["max"], horizontal_packager.wrapper_limits,
                                                             self._platform.max_wallclock).build_vertical_package(job)
            current_package.append(list(set(job_list)))

        for job in current_package[-1]:
            total_wallclock = sum_str_hours(total_wallclock, job.wallclock)
        if len(current_package) > 1:
            for level in range(1, len(current_package)):
                for job in current_package[level]:
                    job.level = level
        return JobPackageVerticalHorizontal(current_package, total_processors, total_wallclock,
                                            jobs_resources=jobs_resources, method=self.wrapper_method[self.current_wrapper_section], configuration=self._as_config, wrapper_section=self.current_wrapper_section )

#TODO rename and unite JobPackerVerticalMixed to JobPackerVertical since the difference between the two is not needed anymore
class JobPackagerVertical(object):
    """
    Vertical Packager Parent Class

    :param jobs_list: Usually there is only 1 job in this list. \n
    :type jobs_list: List() of Job Objects \n
    :param total_wallclock: Wallclock per object. \n
    :type total_wallclock: String  \n
    :param max_jobs: Maximum number of jobs per platform. \n
    :type max_jobs: Integer \n
    :param wrapper_limits: All wrapper limitations are inside this dictionary ( min,max,by_section,horizontal and vertical). \n
    :type wrapper_limits: Dict. \n
    :param max_wallclock: Value from Platform. \n
    :type max_wallclock: Integer

    """

    def __init__(self, jobs_list, total_wallclock, max_jobs, wrapper_limits, max_wallclock, wrapper_info):
        self.jobs_list = jobs_list
        self.total_wallclock = total_wallclock
        self.max_jobs = max_jobs
        self.wrapper_limits = wrapper_limits
        self.max_wallclock = max_wallclock
        self.wrapper_info = wrapper_info

    def build_vertical_package(self, job, level=1):
        """
        Goes through the job and all the related jobs (children, or part of the same date member ordered group), finds those suitable
        and groups them together into a wrapper. 

        :param level:
        :param job: Job to be wrapped. \n
        :type job: Job Object \n
        :return: List of jobs that are wrapped together. \n
        :rtype: List() of Job Object \n
        """
        # self.jobs_list starts as only 1 member, but wrapped jobs are added in the recursion
        if len(self.jobs_list) >= self.wrapper_limits["max_v"] or len(self.jobs_list) >= self.wrapper_limits["max_by_section"][job.section] or len(self.jobs_list) >= self.wrapper_limits["max"]:
            return self.jobs_list
        child = self.get_wrappable_child(job)
        # If not None, it is wrappable
        if child is not None and len(str(child)) > 0:
            # Calculate total wallclock per possible wrapper
            self.total_wallclock = sum_str_hours(
                self.total_wallclock, child.wallclock)
            # Testing against max from platform
            if self.total_wallclock <= self.max_wallclock:
                # Marking, this is later tested in the main loop
                child.packed = True
                child.level = level
                self.jobs_list.append(child)
                # Recursive call
                return self.build_vertical_package(child, level=level + 1)
        # Wrapped jobs are accumulated and returned in this list
        return self.jobs_list

    def get_wrappable_child(self, job):
        pass

    def _is_wrappable(self, job):
        """
        Determines if a job is wrappable. Basically, the job shouldn't have been packed already and the status must be READY or WAITING,
        Its parents should be COMPLETED.

        :param job: job to be evaluated. \n
        :type job: Job Object \n
        :return: True if wrappable, False otherwise. \n
        :rtype: Boolean
        """
        if job.packed is False and (job.status == Status.READY or job.status == Status.WAITING):
            for parent in job.parents:
                # First part of this conditional is true only if the parent is already on the wrapper package ( job_lists == current_wrapped jobs there )
                # Second part is actually relevant, parents of a wrapper should be COMPLETED
                if parent not in self.jobs_list and parent.status != Status.COMPLETED:
                    return False
            return True
        return False

class JobPackagerVerticalMixed(JobPackagerVertical):
    """
    Vertical Mixed Class. First statement of the constructor builds JobPackagerVertical.

    :param dict_jobs: Jobs sorted by date, member, RUNNING, and chunk number. Only those relevant to the wrapper. \n
    :type dict_jobs: Dictionary Key: date, Value: (Dictionary Key: Member, Value: List of jobs sorted) \n
    :param ready_job: Job to be wrapped. \n
    :type ready_job: Job Object \n
    :param jobs_list: ready_job as a list. \n
    :type jobs_list: List() of Job Object \n
    :param total_wallclock: wallclock time per job. \n
    :type total_wallclock: String \n
    :param max_jobs: Maximum number of jobs per platform. \n
    :type max_jobs: Integer \n
    :param wrapper_limits: All wrapper limitations are inside this dictionary ( min,max,by_section,horizontal and vertical). \n
    :type wrapper_limits: Dict. \n
    :param max_wallclock: Value from Platform. \n
    :type max_wallclock: String \n
    """

    def __init__(self, dict_jobs, ready_job, jobs_list, total_wallclock, max_jobs, wrapper_limits, max_wallclock,wrapper_info={}):
        super(JobPackagerVerticalMixed, self).__init__(
            jobs_list, total_wallclock, max_jobs, wrapper_limits, max_wallclock, wrapper_info)
        self.ready_job = ready_job
        self.dict_jobs = dict_jobs
        # Last date from the ordering
        date = list(dict_jobs.keys())[-1]
        # Last member from the last date from the ordering
        member = list(dict_jobs[date].keys())[-1]
        # If job to be wrapped has date and member, use those
        if ready_job.date is not None and len(str(ready_job.date)) > 0:
            date = ready_job.date
        if ready_job.member is not None and len(str(ready_job.member)) > 0:
            member = ready_job.member
        # Extract list of sorted jobs per date and member
        self.sorted_jobs = dict_jobs[date][member]
        self.index = 0


    def get_wrappable_child(self, job):
        """
        Goes through the jobs with the same date and member than the input job, and return the first that satisfies self._is_wrappable()

        :param job: job to be evaluated. \n
        :type job: Job Object \n
        :return: job that is wrappable. \n
        :rtype: Job Object
        """
        sorted_jobs = self.sorted_jobs

        for index in range(self.index, len(sorted_jobs)):
            child = sorted_jobs[index]
            if self._is_wrappable(child):
                self.index = index + 1
                return child
            continue
        return None
        # Not passing tests but better wrappers result to check
        # for child in job.children:
        #     if child.name != job.name:
        #         if self._is_wrappable(child):
        #             self.index = self.index + 1
        #             return child
        #     continue
        # return None

    def _is_wrappable(self, job):
        """
        Determines if a job is wrappable. Basically, the job shouldn't have been packed already and the status must be READY or WAITING,
        Its parents should be COMPLETED.

        :param job: job to be evaluated. \n
        :type job: Job Object \n
        :return: True if wrappable, False otherwise. \n
        :rtype: Boolean
        """
        if job.packed is False and (job.status == Status.READY or job.status == Status.WAITING):
            for parent in job.parents:
                # First part of this conditional is true only if the parent is already on the wrapper package ( job_lists == current_wrapped jobs there )
                # Second part is actually relevant, parents of a wrapper should be COMPLETED
                if parent not in self.jobs_list and parent.status != Status.COMPLETED:
                    return False
            return True
        return False


class JobPackagerHorizontal(object):
    def __init__(self, job_list, max_processors, wrapper_limits, max_jobs, processors_node, method="ASThread"):
        self.processors_node = processors_node
        self.max_processors = max_processors
        self.wrapper_limits = wrapper_limits
        self.job_list = job_list
        self.max_jobs = max_jobs
        self._current_processors = 0
        self._sort_order_dict = dict()
        self._components_dict = dict()
        self._section_processors = dict()
        self.method = method

        self._maxTotalProcessors = 0
        self._sectionList = list()
        self._package_sections = dict()

    def build_horizontal_package(self, horizontal_vertical=False):
        current_package = []
        current_package_by_section = {}
        if horizontal_vertical:
            self._current_processors = 0
        jobs_by_section = dict()
        for job in self.job_list:
            if job.section not in jobs_by_section:
                jobs_by_section[job.section] = list()
            jobs_by_section[job.section].append(job)
        for section in jobs_by_section:
            current_package_by_section[section] = 0
            for job in jobs_by_section[section]:
                if str(job.processors).isdigit() and str(job.nodes).isdigit() and int(job.nodes) > 1 and int(job.processors) <= 1:
                    job.processors = 0
                if job.total_processors == "":
                    job_total_processors = 0
                else:
                    job_total_processors = int(job.total_processors)
                if len(current_package) < self.wrapper_limits["max_h"] and len(current_package) < self.wrapper_limits["max"]  and current_package_by_section[section] < self.wrapper_limits["max_by_section"][section]:
                    if int(job.tasks) != 0 and int(job.tasks) != int(self.processors_node) and \
                            int(job.tasks) < job_total_processors:
                        nodes = int(
                            ceil(job_total_processors / float(job.tasks)))
                        total_processors = int(self.processors_node) * nodes
                    else:
                        total_processors = job_total_processors
                    if (self._current_processors + total_processors) <= int(self.max_processors):
                        current_package.append(job)
                        self._current_processors += total_processors
                    else:
                        current_package = [job]
                        self._current_processors = total_processors
                    current_package_by_section[section] += 1
                else:
                    break

        self.create_components_dict()

        return current_package

    def create_sections_order(self, jobs_sections):
        for i, section in enumerate(jobs_sections.split('&')):
            self._sort_order_dict[section] = i

    # EXIT FALSE IF A SECTION EXIST AND HAVE LESS PROCESSORS
    def add_sectioncombo_processors(self, total_processors_section):
        keySection = ""

        self._sectionList.sort()
        for section in self._sectionList:
            keySection += str(section)
        if keySection in self._package_sections:
            if self._package_sections[keySection] < total_processors_section:
                return False
        else:
            self._package_sections[keySection] = total_processors_section
        self._maxTotalProcessors = max(
            max(self._package_sections.values()), self._maxTotalProcessors)
        return True

    def sort_by_expression(self, jobname):
        jobname = jobname.split('_')[-1]
        return self._sort_order_dict[jobname]

    def get_next_packages(self, jobs_sections, max_wallclock=None, potential_dependency=None, packages_remote_dependencies=list(), horizontal_vertical=False, max_procs=0):
        packages = []
        job = max(self.job_list, key=attrgetter('total_wallclock'))
        wallclock = job.wallclock
        total_wallclock = wallclock

        while self.max_jobs > 0:
            next_section_list = []
            for job in self.job_list:
                for child in job.children:
                    if job.section == child.section or (job.section in jobs_sections and child.section in jobs_sections.split("&")) \
                            and child.status in [Status.READY, Status.WAITING]:
                        wrappable = True
                        for other_parent in child.parents:
                            if other_parent.status != Status.COMPLETED and other_parent not in self.job_list:
                                wrappable = False
                        if wrappable and child not in next_section_list:
                            next_section_list.append(child)

            next_section_list.sort(
                key=lambda job: self.sort_by_expression(job.name))
            self.job_list = next_section_list
            package_jobs = self.build_horizontal_package(horizontal_vertical)

            if package_jobs:
                sections_aux = set()
                wallclock = package_jobs[0].wallclock
                for job in package_jobs:
                    if job.section not in sections_aux:
                        sections_aux.add(job.section)
                        if job.wallclock > wallclock:
                            wallclock = job.wallclock
                if self._current_processors > max_procs:
                    return packages
                if max_wallclock:
                    total_wallclock = sum_str_hours(total_wallclock, wallclock)
                    if total_wallclock > max_wallclock:
                        return packages
                packages.append(package_jobs)

            else:
                break

        return packages

    @property
    def total_processors(self):
        return self._current_processors

    @property
    def components_dict(self):
        return self._components_dict

    def create_components_dict(self):
        self._sectionList = []
        for job in self.job_list:
            if job.section not in self._sectionList:
                self._sectionList.append(job.section)
            if job.section not in self._components_dict:
                self._components_dict[job.section] = dict()
                self._components_dict[job.section]['COMPONENTS'] = {parameter: job.parameters[parameter]
                                                                    for parameter in list(job.parameters.keys())
                                                                    if '_NUMPROC' in parameter}



#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from .platform_monitor import PlatformMonitor
from .slurm_monitor_item import SlurmMonitorItem

class SlurmMonitor(PlatformMonitor):
  """ Manages Slurm commands interpretation. """
  def __init__(self, platform_output):
      super(SlurmMonitor, self).__init__(platform_output)
      self._identify_input_rows()

  @property
  def steps_energy(self):
    return sum([step.energy for step in self.input_items if step.is_step])
    
  @property
  def total_energy(self):
    return max(self.header.energy, self.steps_energy + self.extern.energy)

  @property
  def step_count(self):
    return len([step for step in self.input_items if step.is_step])

  def _identify_input_rows(self):
      lines = self.input.split("\n")
      self.input_items = [SlurmMonitorItem.from_line(line) for line in lines]        
  
  @property
  def steps(self):
    return [item for item in self.input_items if item.is_step]

  @property
  def header(self):
    # test
    headers = [header for header in self.input_items if header.is_header]
    if len(headers) > 0:
        return headers[0]
    else:
        return None
  
  @property
  def batch(self):
    #test
    batch = [batch for batch in self.input_items if batch.is_batch]
    if len(batch) > 0:
        return batch[0]
    else:
        return None

  @property
  def extern(self):
    #test
    extern = [extern for extern in self.input_items if extern.is_extern]
    if len(extern) > 0:
        return extern[0]
    else:
        return None

  def steps_plus_extern_approximate_header_energy(self):
    return abs(self.steps_energy + self.extern.energy - self.header.energy) <= 0.01*self.header.energy

  def print_items(self):
    for item in self.input_items:
      print(item)  





#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from autosubmit.history.utils import get_current_datetime_if_none

class ExperimentRun:
  """
  Class that represents an experiment run
  """
  def __init__(self, run_id, created=None, modified=None, start=0, finish=0, chunk_unit="NA", chunk_size=0, completed=0, total=0, failed=0, queuing=0, running=0, submitted=0, suspended=0, metadata=""):
    self.run_id = run_id
    self.created = get_current_datetime_if_none(created) 
    self.modified = get_current_datetime_if_none(modified) # Added on DB 16
    self.start = start
    self.finish = finish
    self.chunk_unit = chunk_unit
    self.chunk_size = chunk_size
    self.submitted = submitted
    self.queuing = queuing
    self.running = running
    self.completed = completed
    self.failed = failed
    self.total = total
    self.suspended = suspended
    self.metadata = metadata    

  @classmethod
  def from_model(cls, row):
    """ Build ExperimentRun from ExperimentRunRow """
    experiment_run = cls(0)
    experiment_run.run_id = row.run_id
    experiment_run.created = get_current_datetime_if_none(row.created)
    experiment_run.modified = get_current_datetime_if_none(row.modified)
    experiment_run.start = row.start
    experiment_run.finish = row.finish
    experiment_run.chunk_unit = row.chunk_unit
    experiment_run.chunk_size = row.chunk_size
    experiment_run.completed = row.completed
    experiment_run.total = row.total
    experiment_run.failed = row.failed
    experiment_run.queuing = row.queuing
    experiment_run.running = row.running
    experiment_run.submitted = row.submitted
    experiment_run.suspended = row.suspended
    experiment_run.metadata = row.metadata
    return experiment_run
  